import logging

import pytest


logger = logging.getLogger()


@pytest.fixture(autouse=True)
def ensure_logging_framework_not_altered():
    # Fix/work around I/O error when running tests:
    #
    # --- Logging error ---
    # Traceback (most recent call last):
    #   File "/usr/lib/python3.10/logging/__init__.py", line 1103, in emit
    #     stream.write(msg + self.terminator)
    # ValueError: I/O operation on closed file.
    # Call stack:
    #   File "/home/mrw/src/os2mo-data-import-and-export/.venv/lib/python3.10/site-packages/aiohttp/client.py", line 348, in __del__
    #     self._loop.call_exception_handler(context)
    #   File "/usr/lib/python3.10/asyncio/base_events.py", line 1778, in call_exception_handler
    #     self.default_exception_handler(context)
    #   File "/usr/lib/python3.10/asyncio/base_events.py", line 1752, in default_exception_handler
    #     logger.error('\n'.join(log_lines), exc_info=exc_info)
    # Message: 'Unclosed client session\nclient_session: <aiohttp.client.ClientSession object at 0x7f539a729c30>'
    # Arguments: ()
    #
    # Source: https://github.com/pytest-dev/pytest/issues/14#issuecomment-521577819
    before_handlers = list(logger.handlers)
    yield
    logger.handlers = before_handlers



import logging.config
import sys

from .read_ad_conf_settings import read_settings


class PasswordRemovalFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        settings = kwargs.pop("settings")
        super().__init__(*args, **kwargs)
        self._passwords = set(self._get_passwords_from_settings(settings))

    def format(self, record):
        original = logging.Formatter.format(self, record)
        return self._remove_password(original)

    def _get_passwords_from_settings(self, settings):
        for key in ("primary", "global"):
            password = settings.get(key, {}).get("password")
            if password:
                yield password

    def _remove_password(self, s):
        for password in self._passwords:
            s = s.replace(password, "*" * len(password))
        return s


def start_logging(**kwargs):
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "format": "%(levelname)s %(asctime)s %(name)s: %(message)s",
                "()": PasswordRemovalFormatter,
                "settings": kwargs.get("settings") or read_settings(),
            },
        },
        "handlers": {
            "local": {
                "formatter": "default",
                "class": "logging.StreamHandler",
                "stream": sys.stdout,
            },
        },
        "loggers": {
            "": {
                "handlers": ["local"],
                "level": "DEBUG",
            },
            "urllib3": {
                "level": "WARNING",
            },
        },
    }

    logging.config.dictConfig(config)



from ra_utils.generate_uuid import uuid_generator

uuid_gen = uuid_generator("aarhus-los")

AZID_SYSTEM = uuid_gen("AZID_SYSTEM")

PERSON_PHONE = uuid_gen("PERSON_PHONE")
PERSON_EMAIL = uuid_gen("PERSON_EMAIL")
PERSON_ROOM = uuid_gen("PERSON_ROOM")
PERSON_JOB_TITLE_ALT = uuid_gen("PERSON_JOB_TITLE_ALT")

UNIT_LOS = uuid_gen("UNIT_LOS")
UNIT_CVR = uuid_gen("UNIT_CVR")
UNIT_EAN = uuid_gen("UNIT_EAN")
UNIT_PNR = uuid_gen("UNIT_PNR")
UNIT_SENR = uuid_gen("UNIT_SENR")
UNIT_DEBITORNR = uuid_gen("UNIT_DEBITORNR")
UNIT_POSTADDR = uuid_gen("UNIT_POSTADDR")
UNIT_WWW = uuid_gen("UNIT_WWW")
UNIT_HOURS = uuid_gen("UNIT_HOURS")
UNIT_PHONEHOURS = uuid_gen("UNIT_PHONEHOURS")
UNIT_PHONE = uuid_gen("UNIT_PHONE")
UNIT_FAX = uuid_gen("UNIT_FAX")
UNIT_EMAIL = uuid_gen("UNIT_EMAIL")
UNIT_MAG_ID = uuid_gen("UNIT_MAG_ID")
UNIT_NAME_ALT = uuid_gen("UNIT_NAME_ALT")

LINJE_ORG_HIERARCHY = uuid_gen("LINJE_ORG_HIERARCHY")
SIKKERHEDS_ORG_HIERARCHY = uuid_gen("SIKKERHEDS_ORG_HIERARCHY")

NOT_PRIMARY = uuid_gen("NOT_PRIMARY")
PRIMARY = uuid_gen("PRIMARY")

ORG_UUID = uuid_gen("ORG_UUID")



import re
import string
from copy import deepcopy
from itertools import chain
from itertools import repeat
from typing import List
from typing import Optional

from hypothesis import given
from hypothesis import strategies as st

from ..ad_template_engine import template_powershell
from ..ad_writer import ADWriter
from .mocks import MO_UUID
from .mocks import MockMORESTSourcePreview


_base_settings = {
    "primary": {"cpr_separator": ""},
    "primary_write": {
        "cpr_field": "unused_cpr",
        "org_field": "unused_location",
        "uuid_field": "unused_uuid",
        "upn_end": "unused_upn_end",
        "mo_to_ad_fields": {},
    },
}


class _TestableADWriter(ADWriter):
    def __init__(self, location: str) -> None:
        self.datasource = MockMORESTSourcePreview()
        self.settings = self.all_settings = _base_settings
        self._location = location

    def _find_unit_info(self, unit_uuid):
        return {"location": self._location}


class TestLocationElement:
    # Use location elements from this offset in test
    _offset = 2

    _template_to_ad_fields = {
        "department": "{{ mo_values['location']|location_element(2) }}",
        "physicalDeliveryOfficeName": "{{ mo_values['location']|location_element(3) }}",
        "extensionAttribute4": "{{ mo_values['location']|location_element(4) }}",
        "extensionAttribute5": "{{ mo_values['location']|location_element(5) }}",
        "extensionAttribute6": "{{ mo_values['location']|location_element(6) }}",
    }

    @given(
        st.lists(
            st.text(alphabet=string.ascii_letters, min_size=1), min_size=2, max_size=10
        )
    )
    def test_location(self, elems: List[str]) -> None:
        # Construct input variables based on Hypothesis data
        location: str = "\\".join(elems)
        # Construct dictionary of expected field names and values in PowerShell output
        expected_field_names = self._template_to_ad_fields.keys()
        expected_elements = elems[self._offset :]
        # Construct dictionary by zipping keys and values
        expected_pairs = dict(
            zip(
                # Take dictionary keys from `expected_field_names`, which are all the
                # keys in `_template_to_ad_fields`.
                expected_field_names,
                # Take dictionary values from `expected_elements`, padding the list of
                # values with `None` if there are not enough values in
                # `expected_elements` to match every key.
                # Paddding with `None` means that we will test that "unmapped" field
                # names will *not* be written by the generated PowerShell code.
                chain(expected_elements, repeat(None)),
            )
        )
        # Arrange
        ad_writer = _TestableADWriter(location)
        settings = deepcopy(_base_settings)
        settings["primary_write"]["template_to_ad_fields"] = self._template_to_ad_fields  # type: ignore
        # Act
        result = template_powershell(
            context={
                "mo_values": ad_writer.read_ad_information_from_mo(MO_UUID),
                "user_sam": None,  # not used
            },
            settings=settings,
            environment=ad_writer._get_jinja_environment(),
        )
        # Assert
        self._assert_matches(result, expected_pairs)

    def _assert_matches(self, result: str, expected: dict):
        for name, expected_value in expected.items():
            if expected_value:
                actual_value = self._get_value(result, name)
                assert expected_value == actual_value, f"mismatch on '{name}'"
            else:
                assert name not in result

    def _get_value(self, result: str, field_name: str) -> Optional[str]:
        # Match both "-attr=value" and '"attr"="value"' argument styles in PowerShell
        # output `result`.
        for regex in (rf"-{field_name} \"(.*?)\"", rf"{field_name}\"=\"(.*?)\""):
            match: Optional[re.Match] = re.search(regex, result)
            if match:
                return match.groups()[0]
        return None



from unittest import mock
from unittest import TestCase

from parameterized import parameterized

from ..ad_writer import ADWriter
from ..mo_to_ad_sync import run_mo_to_ad_sync
from ..mo_to_ad_sync import run_preview_command_for_uuid
from .mocks import MO_UUID
from .mocks import MockADParameterReader
from .mocks import MockADParameterReaderWithManager
from .mocks import MockADWriterContext
from .mocks import MockLoraCacheExtended
from .mocks import MockLoraCacheWithManager
from .test_utils import dict_modifier
from .test_utils import TestADWriterMixin


@mock.patch("time.sleep")
class TestMoToAdSync(TestCase, TestADWriterMixin):
    """Test `mo_to_ad_sync`"""

    def setUp(self):
        super().setUp()

        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Name": "{{ mo_values['full_name'] }} - testing"
                },
            }
        )

        def remove_manager_cpr(mo_values, *args, **kwargs):
            del mo_values["manager_cpr"]
            return mo_values

        self._setup_adwriter(
            early_transform_settings=settings_transformer,
            transform_mo_values=remove_manager_cpr,
        )
        self._mock_reader = MockADParameterReader()

    def test_all_ad_users(self, *args):
        self._assert_stats_ok(self._run())

    def test_single_ad_user(self, *args):
        self._assert_stats_ok(self._run(sync_username="not important"))

    def test_single_ad_user_can_be_renamed(self, *args):
        self._run(sync_username="not important")
        rename_cmd = self.ad_writer.scripts[0].splitlines()[-1]
        self.assertIn("-NewName", rename_cmd)

    def test_invalid_uuid_field_does_nothing(self, *args):
        self._assert_stats_ok(
            self._run(mo_uuid_field="invalid field name"),
            num_attempted=0,
            num_successful=0,
        )

    def test_unhandled_exception(self, *args):
        with mock.patch.object(self.ad_writer, "sync_user", side_effect=Exception):
            with self.assertLogs("export") as cm:
                self._assert_stats_ok(
                    self._run(),
                    num_successful=0,
                    num_critical_error=1,
                )
                self.assertEqual(len(cm.records), 1)
                self.assertRegex(
                    cm.records[0].message, r"Error updating AD user '.*?': .*"
                )

    def _run(self, mo_uuid_field="ObjectGUID", **kwargs):
        return run_mo_to_ad_sync(
            self._mock_reader,
            self.ad_writer,
            mo_uuid_field=mo_uuid_field,
            **kwargs,
        )

    def _assert_stats_ok(
        self,
        stats,
        num_attempted=1,
        num_successful=1,
        num_critical_error=0,
    ):
        self.assertDictEqual(
            stats,
            {
                "attempted_users": num_attempted,
                "updated": num_successful,
                "fully_synced": num_successful,
                "nothing_to_edit": 0,
                "no_manager": 0,
                "unknown_manager_failure": 0,
                "cpr_not_unique": 0,
                "user_not_in_mo": 0,
                "user_not_in_ad": 0,
                "critical_error": num_critical_error,
                "unknown_failed_sync": 0,
                "no_active_engagement": 0,
            },
        )


class TestMoToAdSyncDryRun:
    @parameterized.expand(
        [
            # 1. Engagement in the present, dry run should count it as an update
            (
                # `lc` contains (default) primary engagement in the present
                MockLoraCacheExtended(mo_values={"uuid": MO_UUID}),
                # Assert that the dry run would count this MO user as one to process
                lambda stats: stats["attempted_users"] == stats["updated"] == 1,
            ),
            # 2. No engagements, dry run should count it as a skipped user
            (
                # `lc` contains no engagements
                MockLoraCacheExtended(mo_values={"uuid": MO_UUID}, mo_engagements=[]),
                # Assert that the dry run would count this MO user as skipped, due to
                # having no active primary engagements.
                lambda stats: stats["attempted_users"] == stats["nothing_to_edit"] == 1,
            ),
        ]
    )
    def test_dry_run_counts_skipped_when_no_engagements_in_loracache(
        self, lc, test_stats
    ):
        # Arrange
        reader = MockADParameterReader()
        with MockADWriterContext() as mock_ad_writer_context:
            writer = ADWriter(lc=lc, lc_historic=lc)
            writer.sync_user = mock.MagicMock()
            # Act
            stats = run_mo_to_ad_sync(reader, writer, "ObjectGUID", dry_run=True)
            # Assert stats are as expected
            assert test_stats(stats)
            # Assert that no actual Powershell commands were issued
            writer.sync_user.assert_not_called()
            mock_ad_writer_context.mock_session.run_ps.assert_not_called()


class TestMoToAdSyncPreview:
    def test_preview_command_for_uuid(self, *args):
        # Arrange
        reader = MockADParameterReaderWithManager()
        lc = MockLoraCacheWithManager()
        with MockADWriterContext():
            writer = ADWriter(lc=lc, lc_historic=lc)
            with mock.patch("click.echo_via_pager") as mock_echo:
                # Act
                commands = run_preview_command_for_uuid(
                    reader,
                    writer,
                    MO_UUID,
                    sync_username="user_sam",
                )
                # Assert
                assert mock_echo.call_count == 4
                assert len(commands) == 4
                for cmd in commands:
                    assert isinstance(cmd, str)



import logging
from datetime import datetime
from functools import partial
from operator import itemgetter
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple

import click
import sentry_sdk
from more_itertools import only
from more_itertools import partition
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.apply import apply
from ra_utils.jinja_filter import create_filters
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm

from .ad_logger import start_logging
from .ad_reader import ADParameterReader
from exporters.sql_export.lora_cache import get_cache as LoraCache

logger = logging.getLogger("AdSyncRead")


# how to check these classes for noobs
# look at :https://os2mo-test.holstebro.dk/service/o/ORGUUID/f/
# It must be addresses, so we find the address thing for employees
# https://os2mo-test.holstebro.dk/service/o/ORGUUID/f/employee_address_type/
# There You have it - for example the mobile phone
# Now You may wonder if the VISIBLE/SECRET are right:
# Find them here https://os2mo-test.holstebro.dk/service/o/ORGUUID/f/visibility/


# AD has no concept of temporality, validity is always from now to infinity.
VALIDITY = {"from": datetime.strftime(datetime.now(), "%Y-%m-%d"), "to": None}


seeded_create_filters = partial(create_filters, tuple_keys=("uuid", "ad_object"))


class ConfigurationError(Exception):
    pass


class AddressDecisionList:
    """Given an AD object and a set of MO adresses, an instance of this class
    represents the updates required to synchronize the MO address data with
    the current state in the AD object.

    Each `AddressDecisionList` instance is an iterable over one or more
    decisions. Each decision is a tuple of `(decision_type, address, *args).`
    Client code can iterate over the list of decisions and effectuate MO API
    calls based on their decision type.

    :param uuid: uuid of the user
    :param ad_object: AD object for user
    :param user_addresses: MO addresses for user
    :param address_mapping: The contents of "user_addresses" in setting
    "ad_mo_sync_mapping"
    :param visibility: The contents of `AdMoSync.visibility`
    """

    # Decision types
    CREATE = "create"
    EDIT = "edit"
    TERMINATE = "terminate"

    def __init__(
        self,
        uuid: str,
        ad_object: dict,
        user_addresses: list,
        address_mapping: dict,
        visibility: dict,
    ):
        self._uuid = uuid
        self._ad_object = ad_object
        self._user_addresses = user_addresses
        self._address_mapping = address_mapping
        self._visibility = visibility
        self._decisions = self._build()

    def __iter__(self) -> Iterator[Tuple[str, Optional[Dict], Optional[Any]]]:
        return self._decisions

    def _build(self):
        for field, (
            address_type_uuid,
            visibility_uuid,
        ) in self._address_mapping.items():
            user_addresses = list(
                filter(
                    partial(self._match_address, address_type_uuid, visibility_uuid),
                    self._user_addresses,
                )
            )

            # No corresponding MO addresses found, let's create one
            if not user_addresses:
                # The field exists in AD, but there's no corresponding MO
                # address.
                if self._ad_object.get(field):
                    # Yield a MO address creation.
                    yield (
                        self.CREATE,
                        None,  # empty MO address
                        self._uuid,
                        self._ad_object[field],
                        (address_type_uuid, visibility_uuid),
                    )
                    continue

            # Field not in AD, terminate all corresponding MO addresses
            if self._ad_object.get(field) is None:
                for address in user_addresses:
                    yield (self.TERMINATE, address)
                continue

            # At this point we know that AD has a value for this field.
            # Only process the MO addresses that differ from AD.
            differing_addresses = filter(
                partial(self._mo_and_ad_differs, field),
                user_addresses,
            )

            # First element in `differing_addresses` is edited, and all other
            # elements are terminated.
            first_address = next(differing_addresses, None)
            if first_address:
                yield (
                    self.EDIT,
                    first_address,
                    self._ad_object[field],
                    (address_type_uuid, visibility_uuid),
                )
            for address in differing_addresses:
                yield (self.TERMINATE, address)

    def _mo_and_ad_differs(self, field: str, address: dict) -> bool:
        value_differs = address["value"] != self._ad_object[field]

        _, visibility_scope = self._address_mapping[field]
        current_visibility_scope = (
            address["visibility"]["scope"] if address.get("visibility") else None
        )
        visibility_differs = visibility_scope != current_visibility_scope

        return value_differs or visibility_differs

    def _match_address(
        self, address_type_uuid: str, visibility_uuid: str, address: dict
    ) -> bool:
        return address is not None and self._match_address_type_uuid(
            address_type_uuid, address
        )

    def _match_address_type_uuid(self, address_type_uuid, address) -> bool:
        # Filter out addresses with wrong type
        return address["address_type"]["uuid"] == address_type_uuid


class AdMoSync:
    def __init__(self, all_settings=None):
        logger.info("AD Sync Started")
        self._setup_settings(all_settings)  # Populates `self.settings`
        self.lc = self._setup_lora_cache()  # Depends on `self.settings`
        self.helper = self._setup_mora_helper()  # Depends on `self.settings`
        self._setup_visibilities()  # Depends on `self.settings` and `self.helper`
        self.org = self.helper.read_organisation()

    def _setup_settings(self, all_settings):
        self.settings = all_settings
        if self.settings is None:
            self.settings = load_settings()

    def _setup_visibilities(self):
        mo_visibilities = self.helper.read_classes_in_facet("visibility")[0]
        self.visibility = {
            "PUBLIC": self.settings["address.visibility.public"],
            "INTERNAL": self.settings["address.visibility.internal"],
            "SECRET": self.settings["address.visibility.secret"],
        }

        # Check that the configured visibilities are found in MO
        configured_visibilities = set(self.visibility.values())
        mo_visibilities = set(map(itemgetter("uuid"), mo_visibilities))
        # If the configured visibiltities are not a subset, at least one is missing.
        if not configured_visibilities.issubset(mo_visibilities):
            raise Exception("Error in visibility class configuration")

    def _setup_mora_helper(self):
        return MoraHelper(hostname=self.settings["mora.base"], use_cache=False)

    def _setup_lora_cache(self):
        # Possibly get IT-system directly from LoRa for better performance.
        lora_speedup = self.settings.get(
            "integrations.ad.ad_mo_sync_direct_lora_speedup", False
        )
        if lora_speedup:
            print("Retrieve LoRa dump")
            lc = LoraCache(resolve_dar=False, full_history=False)
            lc.populate_cache(skip_associations=True, dry_run=False)
            # skip reading lora - not for prod
            # lc.populate_cache(dry_run=True, skip_associations=True)
            lc.calculate_primary_engagements()
            print("Done")
            return lc
        print("Use direct MO access")
        return None

    def _read_all_mo_users(self):
        """Return a list of all employees in MO.

        :return: List af all employees.
        """
        logger.info("Read all MO users")
        if self.lc:
            employees = [
                employee[0] for employee in self.lc.users.values() if len(employee) > 0
            ]
        else:
            employees = self.helper.read_all_users()
        logger.info("Done reading all MO users")
        return employees

    def _read_itconnections_raw(self, uuid, it_system_uuid=None):
        logger.debug(f"Read it-system for user {uuid}")
        if self.lc:
            itconnections = map(itemgetter(0), self.lc.it_connections.values())
            itconnections = filter(lambda it: it["user"] == uuid, itconnections)
            if it_system_uuid:
                itconnections = filter(
                    lambda it: it["itsystem"] == it_system_uuid, itconnections
                )
        else:
            itconnections = self.helper.get_e_itsystems(uuid, it_system_uuid)
        return itconnections

    def _read_itconnections(self, uuid, it_system_uuid=None):
        # Figure out which fields to extract from the it-system
        # Differs by source, as LoraCache and MO are not equivalent!
        extractor = itemgetter("user_key", "uuid")
        if self.lc:
            extractor = itemgetter("username", "uuid")
        # Fetch itconnections and extract fields
        itconnections = self._read_itconnections_raw(uuid, it_system_uuid)
        itconnections = map(extractor, itconnections)
        return itconnections

    def _get_address_decision_list(self, uuid, ad_object):
        """Construct a `AddressDecisionList` instance for `ad_object`

        :param uuid: uuid of the user
        :param ad_object: AD object for user
        :return: An `AddressDecisionList` instance
        """

        def to_mo_address(addr):
            return {
                "uuid": addr["uuid"],
                "address_type": {"uuid": addr["adresse_type"]},
                "visibility": {"uuid": addr["visibility"]},
                "value": addr["value"],
                "validity": {
                    "from": addr["from_date"],
                    "to": addr["to_date"],
                },
            }

        # Populate list of `user_addresses`
        if self.lc:
            # Retrieve user addresses from LoraCache
            user_addresses = self.lc.addresses.values()
            user_addresses = map(itemgetter(0), user_addresses)
            user_addresses = filter(lambda addr: addr["user"] == uuid, user_addresses)
            user_addresses = map(to_mo_address, user_addresses)
            user_addresses = list(user_addresses)
        else:
            # Retrieve user addresses from MO
            user_addresses = self.helper.get_e_addresses(uuid)

        return AddressDecisionList(
            uuid,
            ad_object,
            user_addresses,
            self.mapping["user_addresses"],
            self.visibility,
        )

    def _create_address(self, uuid, value, klasse):
        """Create a new address for a user.

        :param uuid: uuid of the user.
        :param: value Value of of the adress.
        :param: klasse: The address type and vissibility of the address.
        """
        payload = {
            "value": value,
            "address_type": {"uuid": klasse[0]},
            "person": {"uuid": uuid},
            "type": "address",
            "validity": VALIDITY,
            "org": {"uuid": self.org},
        }
        if klasse[1] is not None:
            payload["visibility"] = {"uuid": self.visibility[klasse[1]]}
        logger.debug("Create payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        logger.debug("Response: {}".format(response.text))

    def _edit_address(self, address_uuid, value, klasse, validity=VALIDITY):
        """Edit an exising address to a new value.

        :param address_uuid: uuid of the address object.
        :param value: The new value
        :param: klasse: The address type and vissibility of the address.
        """
        payload = [
            {
                "type": "address",
                "uuid": address_uuid,
                "data": {
                    "validity": validity,
                    "value": value,
                    "address_type": {"uuid": klasse[0]},
                },
            }
        ]
        if klasse[1] is not None:
            payload[0]["data"]["visibility"] = {"uuid": self.visibility[klasse[1]]}

        logger.debug("Edit payload: {}".format(payload))
        response = self.helper._mo_post("details/edit", payload)
        logger.debug("Response: {}".format(response.text))

    def _edit_engagement(self, uuid: str, ad_object):
        if "engagements" not in self.mapping:
            return

        def _make_exclude_function(fieldspec):
            """Given a callable `fieldspec`, return a function which will
            return True if the date found by `fieldspec` is before today, and
            otherwise will return False.

            This is used to exclude engagements with a "from date" in the
            future.
            """

            threshold = datetime.now().date()

            def _func(engagement):
                value = fieldspec(engagement)
                if value is None:
                    return False  # from_date is infinitely far in the future
                if isinstance(value, str):
                    value = datetime.strptime(value, "%Y-%m-%d").date()
                return value < threshold

            return _func

        if self.lc:
            # Read user's current engagements, e.g. exclude engagements that
            # ended in the past.
            engagements = self.lc.engagements.values()
            engagements = map(itemgetter(0), engagements)
            engagements = filter(lambda eng: eng["user"] == uuid, engagements)
            # Skip engagements beginning in the future
            engagements = filter(
                _make_exclude_function(itemgetter("from_date")),
                list(engagements),
            )
            # Notice, this will only update current row, if more rows exists, they
            # will not be updated until the first run after that row has become
            # current. To fix this, we will need to ad option to LoRa cache to be
            # able to return entire object validity (poc-code exists).
        else:
            # Read user's current engagements, e.g. exclude engagements that
            # ended in the past.
            engagements = self.helper.read_user_engagement(
                uuid,
                calculate_primary=True,
                read_all=True,
                skip_past=True,
            )
            # Skip engagements beginning in the future
            engagements = filter(
                _make_exclude_function(lambda eng: eng["validity"]["from"]),
                list(engagements),
            )

        for engagement in engagements:
            if self.lc:
                ad_object_compare = ad_object if engagement["primary_boolean"] else {}
            else:
                ad_object_compare = ad_object if engagement["is_primary"] else {}

            to_date = (
                engagement["to_date"]
                if self.lc
                else engagement.get("validity", {}).get("to")
            )
            validity = {
                "from": VALIDITY["from"],  # today
                "to": to_date if to_date != "9999-12-31" else None,
            }
            self._edit_engagement_post_to_mo(
                uuid, ad_object_compare, engagement, validity
            )

    def _edit_engagement_post_to_mo(self, uuid, ad_object, mo_engagement, validity):
        # Populate `mo_data` with a value for each mapped field whose value has
        # changed when comparing the AD object to the current MO object.
        mo_data = {}
        for ad_field, mo_field in self.mapping["engagements"].items():
            # Default `mo_value` to an empty string. In case the field is
            # dropped from the AD object, this will empty its value in MO.
            new_mo_value = ad_object.get(ad_field, None)
            old_mo_value = mo_engagement.get(mo_field)

            # If we cannot read the field, maybe it is because our
            # mo_engagement is from LoraCache, and thus is different from MO
            # and must be read differently.
            if old_mo_value is None and "extensions" in mo_engagement:
                old_mo_value = self._edit_engagement_read_lc_extensions(
                    mo_field,
                    mo_engagement,
                )

            if old_mo_value == new_mo_value:
                logger.debug("No change, not editing engagement")
                continue

            mo_data[mo_field] = new_mo_value

        # Check that the loop added at least one key/value pair to `mo_data`.
        if not mo_data:
            logger.debug("Nothing to post, returning")
            return

        # Then, add the validity.
        mo_data["validity"] = validity

        payload = {
            "type": "engagement",
            "uuid": mo_engagement["uuid"],
            "data": mo_data,
        }
        logger.debug("Edit payload: %r", payload)
        response = self.helper._mo_post("details/edit", payload)
        self.stats["engagements"] += 1
        self.stats["users"].add(uuid)
        logger.debug("Response: %r", response.text)

    def _edit_engagement_read_lc_extensions(self, mo_field, mo_engagement):
        # TODO: This is specific to the LoraCache-based implementation (when
        # `self.lc` is not None.) Refactor to use a shared class for reading
        # engagements from LoraCache.
        field_mapping = {
            f"extension_{x}": mo_engagement["extensions"][f"udvidelse_{x}"]
            for x in range(1, 11)
        }
        if mo_field not in field_mapping:
            raise ConfigurationError("MO field %r is not mapped" % mo_field)
        return field_mapping[mo_field]

    def _create_it_system(self, person_uuid, ad_username, mo_itsystem_uuid):
        payload = {
            "type": "it",
            "user_key": ad_username,
            "itsystem": {"uuid": mo_itsystem_uuid},
            "person": {"uuid": person_uuid},
            "validity": VALIDITY,
        }
        logger.debug("Create it system payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        logger.debug("Response: {}".format(response.text))
        response.raise_for_status()

    def _update_it_system(self, ad_username, binding_uuid):
        payload = {
            "type": "it",
            "data": {"user_key": ad_username, "validity": VALIDITY},
            "uuid": binding_uuid,
        }
        logger.debug("Update it system payload: {}".format(payload))
        response = self.helper._mo_post("details/edit", payload)
        logger.debug("Response: {}".format(response.text))
        response.raise_for_status()

    def _edit_it_system(self, uuid, ad_object):
        mo_itsystem_uuid = self.mapping["it_systems"]["samAccountName"]
        itconnections = self._read_itconnections(uuid, mo_itsystem_uuid)
        # Here it_systems is a 2 tuple (mo_username, binding_uuid)
        mo_username, binding_uuid = only(itconnections, ("", ""))
        # Username currently in AD
        ad_username = ad_object["SamAccountName"]

        # If mo_username is blank, we found a user who needs a new entry created
        if mo_username == "":
            self._create_it_system(uuid, ad_username, mo_itsystem_uuid)
            self.stats["it_systems"] += 1
            self.stats["users"].add(uuid)
        elif mo_username != ad_username:  # We need to update the mo_username
            self._update_it_system(ad_username, binding_uuid)
            self.stats["it_systems"] += 1
            self.stats["users"].add(uuid)

    def _edit_user_addresses(self, uuid, ad_object):
        decision_list = self._get_address_decision_list(uuid, ad_object)
        for decision, address, *args in decision_list:
            if decision == AddressDecisionList.CREATE:
                self._create_address(*args)
                # Update internal stats
                self.stats["addresses"][0] += 1
                self.stats["users"].add(uuid)
            elif decision == AddressDecisionList.EDIT:
                self._edit_address(address["uuid"], *args)
                # Update internal stats
                self.stats["addresses"][1] += 1
                self.stats["users"].add(uuid)
            elif decision == AddressDecisionList.TERMINATE:
                self._finalize_user_addresses_post_to_mo(address)
            else:
                raise ValueError(
                    "unknown decision %r (address=%r, args=%r)"
                    % (decision, address, args)
                )

    def _finalize_it_system(self, uuid):
        if "it_systems" not in self.mapping:
            return

        # Figure out how to find the itsystems uuid
        # Differs by source, as LoraCache and MO are not equivalent!
        def itsystem_uuid_extractor_mo(it):
            return it["itsystem"]["uuid"]

        def itsystem_uuid_extractor_lora(it):
            return it["itsystem"]

        itsystem_uuid_extractor = itsystem_uuid_extractor_mo
        if self.lc:
            itsystem_uuid_extractor = itsystem_uuid_extractor_lora

        it_system_uuids = self.mapping["it_systems"].values()

        def _keep_only_mapped(itconnection):
            itsystem_uuid = itsystem_uuid_extractor(itconnection)
            return itsystem_uuid in it_system_uuids

        def _keep_only_open(itconnection):
            # Figure out how to find the itsystem connection end-date
            # Differs by source, as LoraCache and MO are not equivalent!
            if self.lc:
                to_date = itconnection["to_date"]
            else:
                to_date = itconnection["validity"]["to"]
            return to_date is None

        itconnections = self._read_itconnections_raw(uuid)
        itconnections = filter(_keep_only_mapped, itconnections)
        itconnections = filter(_keep_only_open, itconnections)
        itconnections = map(itemgetter("uuid"), itconnections)

        today = datetime.strftime(datetime.now(), "%Y-%m-%d")
        for uuid in itconnections:
            payload = {
                "type": "it",
                "uuid": uuid,
                "validity": {"to": today},
            }
            logger.debug("Finalize payload: {}".format(payload))
            response = self.helper._mo_post("details/terminate", payload)
            logger.debug("Response: {}".format(response.text))

    def _finalize_user_addresses(self, uuid, ad_object):
        if "user_addresses" not in self.mapping:
            return

        @apply
        def _is_terminate(decision, address, *args):
            return decision == AddressDecisionList.TERMINATE

        @apply
        def _has_no_end_date(decision, address, *args):
            return address["validity"]["to"] is None

        @apply
        def _extract_address(decision, address, *args):
            return address

        decision_list = list(self._get_address_decision_list(uuid, ad_object))
        decision_list = filter(_is_terminate, decision_list)
        decision_list = filter(_has_no_end_date, decision_list)
        decision_list = map(_extract_address, decision_list)

        for address in decision_list:
            self._finalize_user_addresses_post_to_mo(address)

    def _finalize_user_addresses_post_to_mo(self, mo_address: dict):
        today = datetime.strftime(datetime.now(), "%Y-%m-%d")
        payload = {
            "type": "address",
            "uuid": mo_address["uuid"],
            "validity": {"to": today},
        }
        logger.debug("Finalize payload: {}".format(payload))
        response = self.helper._mo_post("details/terminate", payload)
        logger.debug("Response: {}".format(response.text))
        return response

    def _edit_user_attrs(self, employee, ad_object):
        user_attrs_mapping = self.mapping.get("user_attrs", {}).items()
        user_attrs_changed = {
            mo_field_name: ad_object.get(ad_field_name)
            for ad_field_name, mo_field_name in user_attrs_mapping
            if (
                ad_object.get(ad_field_name) is not None
                and ad_object[ad_field_name] != employee.get(mo_field_name)
            )
        }
        if user_attrs_changed:
            user_attrs_changed["validity"] = VALIDITY
            self.stats["users"].add(employee["uuid"])
            return self.helper.update_user(employee["uuid"], user_attrs_changed)

    def _terminate_single_user(self, uuid: str, ad_object: dict):
        self._finalize_it_system(uuid)
        self._finalize_user_addresses(uuid, ad_object)
        self._edit_engagement(uuid, {})

    def _update_single_user(
        self,
        employee,
        ad_object,
        terminate_disabled,
        terminate_disabled_filters,
    ):
        """Update all fields for a single user.

        :param employee: MO employee object (dict)
        :param ad_object: Dict with the AD information for the user.
        :param terminate_disabled: terminate *all* disabled users if True
        :param terminate_disabled_filters: list of filter functions
        """
        # If `user_attrs` is set up, transfer values from mapped AD user fields
        # to corresponding MO user attributes.
        if self.mapping.get("user_attrs"):
            self._edit_user_attrs(employee, ad_object)

        # Debug log if enabled is not found
        if "Enabled" not in ad_object:
            logger.info("Enabled not in ad_object")
        user_enabled = ad_object.get("Enabled", True)

        # If terminate_disabled is None, we decide on a per-user basis using the
        # terminate_disabled_filters, by invariant we at least one exist.
        if terminate_disabled is None:
            terminate_disabled = all(
                terminate_disabled_filter((employee["uuid"], ad_object))
                for terminate_disabled_filter in terminate_disabled_filters
            )

        # Check whether the current user is disabled, and terminate them, if we are
        # configured to terminate disabled users.
        if terminate_disabled and not user_enabled:
            # Set validity end --> today if in the future
            self._terminate_single_user(employee["uuid"], ad_object)
            return

        # Sync the user, whether disabled or not
        if "it_systems" in self.mapping:
            self._edit_it_system(employee["uuid"], ad_object)

        if "engagements" in self.mapping:
            self._edit_engagement(employee["uuid"], ad_object)

        if "user_addresses" in self.mapping:
            self._edit_user_addresses(employee["uuid"], ad_object)

    def _setup_ad_reader_and_cache_all(self, index, cache_all=True):
        ad_reader = ADParameterReader(index=index)
        print("Retrieve AD dump")
        if cache_all:
            ad_reader.cache_all(print_progress=True)
        print("Done")
        logger.info("Done with AD caching")
        return ad_reader

    def _verify_it_systems(self):
        """Verify that all configured it-systems exist."""
        if "it_systems" not in self.mapping:
            return

        # Set of UUIDs of all it_systems in MO
        mo_it_systems = set(map(itemgetter("uuid"), self.helper.read_it_systems()))

        @apply
        def filter_found(it_system, it_system_uuid):
            return it_system_uuid not in mo_it_systems

        # List of tuples (name, uuid) of it-systems configured in settings
        configured_it_systems = self.mapping["it_systems"].items()
        # Remove all the ones that exist in MO
        configured_it_systems = filter(filter_found, configured_it_systems)

        for it_system, it_system_uuid in configured_it_systems:
            msg = "{} with uuid {}, not found in MO"
            raise Exception(msg.format(it_system, it_system_uuid))

    def update_single_user(self, uuid):
        employees = [self.helper.read_user(uuid)]
        self._update_users(employees, ad_cache_all=False)

    def update_all_users(self):
        employees = self._read_all_mo_users()
        self._update_users(employees)

    def _update_users(self, employees_in, ad_cache_all=True):
        # Iterate over all AD's
        for index, _ in enumerate(self.settings["integrations.ad"]):
            employees = employees_in

            self.stats = {
                "ad-index": index,
                "addresses": [0, 0],
                "engagements": 0,
                "it_systems": 0,
                "users": set(),
            }

            ad_reader = self._setup_ad_reader_and_cache_all(
                index=index, cache_all=ad_cache_all
            )
            ad_settings = ad_reader._get_setting()

            # move to read_conf_settings og valider på tværs af alle-ad'er
            # så vi ikke overskriver addresser, itsystemer og extensionfelter
            # fra et ad med  med værdier fra et andet
            self.mapping = ad_settings["ad_mo_sync_mapping"]
            self._verify_it_systems()

            used_mo_fields = []

            for key in self.mapping.keys():
                for ad_field, mo_combi in self.mapping.get(key, {}).items():
                    if mo_combi in used_mo_fields:
                        msg = "MO field {} used more than once"
                        raise Exception(msg.format(mo_combi))
                    used_mo_fields.append(mo_combi)

            def add_employee_cpr(employee):
                """Convert an employee to a tuple (cpr, employee)."""
                if "cpr" in employee:
                    cpr = employee["cpr"]
                else:
                    uuid = employee["uuid"]
                    user = self.helper.read_user(uuid)
                    cpr = user.get("cpr_no")
                    if not cpr:
                        logger.warning("no 'cpr_no' for MO user %r", uuid)
                return cpr, employee

            @apply
            def cpr_uuid_to_uuid_ad(cpr, employee):
                ad_object = ad_reader.read_user(cpr=cpr, cache_only=ad_cache_all)
                return employee, ad_object

            @apply
            def filter_no_ad_object(employee, ad_object):
                return ad_object

            # Lookup filter jinja templates
            pre_filters = seeded_create_filters(ad_settings["ad_mo_sync_pre_filters"])
            terminate_disabled_filters = seeded_create_filters(
                ad_settings["ad_mo_sync_terminate_disabled_filters"]
            )
            # Lookup whether or not to terminate missing users
            terminate_missing = ad_settings["ad_mo_sync_terminate_missing"]
            # Decide whether missing users should only be terminated if and only if
            # they have an AD it system in their MO account.
            terminate_missing_require_itsystem = ad_settings[
                "ad_mo_sync_terminate_missing_require_itsystem"
            ]
            # Lookup whether or not to terminate disabled users
            terminate_disabled = ad_settings["ad_mo_sync_terminate_disabled"]

            # If not globally configured, and no user filters are configured either,
            # we default terminate_disabled to False
            if terminate_disabled is None and not terminate_disabled_filters:
                terminate_disabled = False

            # Iterate over all users and sync AD informations to MO.
            employees = map(add_employee_cpr, employees)
            employees = map(cpr_uuid_to_uuid_ad, employees)

            # Remove all entries without ad_object
            missing_employees, employees = partition(filter_no_ad_object, employees)

            # Run all pre filters
            for pre_filter in pre_filters:
                employees = filter(pre_filter, employees)

            # Call update_single_user on each remaining users
            print("Updating users")
            employees = list(employees)
            employees = tqdm(employees)
            for employee, ad_object in employees:
                # TODO: Convert this function into two seperate phases.
                # 1. A map from uuid, ad_object to mo_endpoints + mo_payloads
                # 2. Bulk updating of MO using the data from 1.
                self._update_single_user(
                    employee, ad_object, terminate_disabled, terminate_disabled_filters
                )

            # Call terminate on each missing user
            if terminate_missing:
                print("Terminating missing users")

                @apply
                def has_it_system(uuid, ad_object):
                    mo_itsystem_uuid = self.mapping["it_systems"]["samAccountName"]
                    itconnections = self._read_itconnections(uuid, mo_itsystem_uuid)
                    mo_username, _ = only(itconnections, ("", ""))
                    return mo_username != ""

                if terminate_missing_require_itsystem:
                    missing_employees = filter(has_it_system, missing_employees)

                missing_employees = list(missing_employees)
                missing_employees = tqdm(missing_employees)

                # TODO: Convert this function into two seperate phases.
                # 1. A map from uuid, ad_object to mo_endpoints + mo_payloads
                # 2. Bulk updating of MO using the data from 1.
                for mo_object, ad_object in missing_employees:
                    self._terminate_single_user(mo_object["uuid"], ad_object)

            logger.info("Stats: {}".format(self.stats))


@click.command()
@click.option(
    "--sync-user",
    help="Sync a single user.",
    type=click.UUID,
)
def sync(sync_user):
    sync = AdMoSync()

    if "crontab.SENTRY_DSN" in sync.settings:
        sentry_sdk.init(dsn=sync.settings["crontab.SENTRY_DSN"])

    if sync_user:
        sync.update_single_user(str(sync_user))
    else:
        sync.update_all_users()


if __name__ == "__main__":
    start_logging()
    sync()



import contextlib
import typing
from datetime import datetime
from operator import itemgetter
from typing import Optional
from unittest.mock import MagicMock
from unittest.mock import patch

import aiohttp.client
import los_files
import pytest
import pytest_aioresponses
import util
from aiohttp import ClientResponseError
from aiohttp import ClientSession
from hypothesis import given
from hypothesis import strategies as st
from more_itertools import one
from unittest import mock

from .helpers import mock_config
from los_files import SFTPFileSet

from parameterized import parameterized


class TestParseFilenames:
    def test_parses_prefix_correctly(self):
        with mock_config(import_csv_folder=None, ):
            with mock.patch("los_files.get_fileset_implementation") as mock_get_fileset:
                mock_get_fileset.return_value = los_files.SFTPFileSet()
                with mock.patch.object(SFTPFileSet, "get_modified_datetime",
                                       return_value=datetime(2020, 5, 1, 0, 0, 0)):
                    filenames = [
                        "xxxxOrg_nye_20210128_123742.csv",
                        "Org_nye_20210128_123742.csv",
                        "garbagebagasdasdads",
                    ]

                    expected_filename = "Org_nye_20210128_123742.csv"

                    actual = los_files.parse_filenames(filenames, "Org_nye",
                                                       datetime.min)

                    actual_filename, _ = one(actual)

                    assert expected_filename == actual_filename

    @parameterized.expand(
        [
            # (filename, modified_date, expected_found),
            ("Org_nye_20200128_123742.csv", datetime(2020, 1, 28, 12, 37, 42),
             datetime(2020, 1, 28, 12, 37, 42), False),
            ("Org_nye_20201028_123742.csv", datetime(2020, 10, 28, 12, 37, 42),
             datetime(2020, 10, 28, 12, 37, 42), True),
            ("Org_nye_20210128_123742.csv", datetime(2021, 1, 28, 12, 37, 42),
             datetime(2021, 1, 28, 12, 37, 42), True),
            ("Org_nye_20210128_123742.csv", datetime(2020, 1, 28, 12, 37, 42),
             datetime(2021, 1, 28, 12, 37, 42), False),
            ("Org_nye_20200128_123742.csv", datetime(2021, 1, 28, 12, 37, 42),
             datetime(2020, 1, 28, 12, 37, 42), True),
        ]
    )
    def test_filters_dates_correctly(self, filename: str, modified_date: datetime,
                                     expected_date: datetime,
                                     expected_found: bool):
        with mock_config(import_csv_folder=None):
            with mock.patch("los_files.get_fileset_implementation") as mock_get_fileset:
                mock_get_fileset.return_value = los_files.SFTPFileSet()
                with mock.patch.object(SFTPFileSet, "get_modified_datetime",
                                       return_value=modified_date):
                    expected = []

                    if expected_found:
                        expected = [(filename, expected_date)]

                    actual = los_files.parse_filenames(
                        [filename, ], "Org_nye", datetime(2020, 6, 1, 0, 0, 0)
                    )

                    assert expected == actual

    @given(
        st.permutations(
            [
                "Org_nye_20210101_000000.csv",
                "Org_nye_20200101_000000.csv",
                "Org_nye_20190101_000000.csv",
                "Org_nye_20180101_000000.csv",
            ]
        )
    )
    def test_sorts_output(self, filenames):
        with mock_config(import_csv_folder=None):
            with mock.patch("los_files.get_fileset_implementation") as mock_get_fileset:
                mock_get_fileset.return_value = los_files.SFTPFileSet()
                with mock.patch.object(SFTPFileSet, "get_modified_datetime",
                                       return_value=datetime(2020, 5, 1, 0, 0, 0)):
                    expected_dates = [
                        datetime(2018, 1, 1),
                        datetime(2019, 1, 1),
                        datetime(2020, 1, 1),
                        datetime(2021, 1, 1),
                    ]

                    actual = los_files.parse_filenames(filenames, "Org_nye",
                                                       datetime.min)
                    actual_dates = list(map(itemgetter(1), actual))

                    assert expected_dates == actual_dates


@contextlib.asynccontextmanager
async def _mock_mo_response(
        aioresponses: pytest_aioresponses,
        mo_endpoint: str,
        mo_http_status: int,
        mo_response: Optional[dict] = None,
) -> typing.AsyncGenerator[ClientSession, None]:
    mora_base = "http://example.com:8080"
    # Mock a response from MO
    aioresponses.post(
        f"{mora_base}{mo_endpoint}?force=1",
        status=mo_http_status,
        payload=mo_response or {},
    )
    # Create client session on mock base URL
    with mock_config(
            mora_base=mora_base,
            max_concurrent_requests=1,
            os2mo_chunk_size=1,
    ):
        async with util.get_client_session() as session:
            yield session


@pytest.mark.asyncio
async def test_terminate_details_handles_404_response(
        aioresponses: pytest_aioresponses,
):
    async def _run_test(ignored_http_statuses):
        # Mock a 404 response from MO "terminate" API, and run `terminate_details`
        async with _mock_mo_response(
                aioresponses, "/service/details/terminate", 404
        ) as client_session:
            detail_payloads = [{"foo": "bar"}]
            await util.terminate_details(
                client_session,
                detail_payloads,
                ignored_http_statuses=ignored_http_statuses,
            )

    # Expect no exceptions if 404 is ignored
    await _run_test(ignored_http_statuses=(404,))

    # Expect an exception if `raise_for_status` is called in `submit_payloads`
    with pytest.raises(ClientResponseError):
        await _run_test(ignored_http_statuses=None)


@pytest.mark.asyncio
async def test_unhandled_status_400(aioresponses: pytest_aioresponses):
    async with _mock_mo_response(
            aioresponses, "/service/details/create", 400, {}
    ) as client_session:
        with pytest.raises(ClientResponseError):
            await util.create_details(client_session, [{"foo": "bar"}])


@pytest.mark.asyncio
async def test_handled_status_400(aioresponses: pytest_aioresponses):
    async with _mock_mo_response(
            aioresponses,
            "/service/details/create",
            400,
            {"error_key": "V_DUPLICATED_IT_USER"},
    ) as client_session:
        await util.create_details(client_session, [{"foo": "bar"}])


@pytest.mark.asyncio
@patch("aiohttp.client.ClientSession.post")
async def test_connection_error(
        mock_post: MagicMock, aioresponses: pytest_aioresponses
):
    mock_post.side_effect = aiohttp.client.ServerDisconnectedError
    async with _mock_mo_response(
            aioresponses, "/service/details/create", 500, {}
    ) as client_session:
        with pytest.raises(aiohttp.ServerDisconnectedError):
            await util.create_details(client_session, [{"foo": "bar"}])

        assert mock_post.call_count == util.retry_attempts



import asyncio
import logging
import uuid
from datetime import date
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import config
import los_files
import mox_helpers.payloads as mox_payloads
import pydantic
import uuids
from aiohttp import ClientResponseError
from aiohttp.http_exceptions import HttpBadRequest
from gql import gql
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import ElementNotFound
from os2mo_data_import.mox_data_types import Facet
from pydantic import Field
from raclients.graph.client import GraphQLClient

logger = logging.getLogger(__name__)


class StamCSV(pydantic.BaseModel):
    @staticmethod
    def get_filename() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @staticmethod
    def get_facet_bvn() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def bvn(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def title(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def class_uuid(self) -> uuid.UUID:
        raise NotImplementedError("must be implemented by subclass")


class Engagementstype(StamCSV):
    engagement_type_uuid: uuid.UUID = Field(alias="EngagementstypeUUID")
    engagement_type: str = Field(alias="Engagementstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Engagementstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_type"

    @property
    def bvn(self) -> str:
        return self.engagement_type

    @property
    def title(self) -> str:
        return self.engagement_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.engagement_type_uuid


class Enhedstype(StamCSV):
    org_unit_type_uuid: uuid.UUID = Field(alias="EnhedstypeUUID")
    org_unit_type: str = Field(alias="Enhedstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Enhedstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "org_unit_type"

    @property
    def bvn(self) -> str:
        return self.org_unit_type

    @property
    def title(self) -> str:
        return self.org_unit_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.org_unit_type_uuid


class ITSystem(StamCSV):
    it_system_uuid: uuid.UUID = Field(alias="ITSystemUUID")
    name: str = Field(alias="Name")
    user_key: str = Field(alias="Userkey")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_ITSystem.csv"

    @property
    def title(self) -> str:
        return self.name

    @property
    def bvn(self) -> str:
        return self.user_key

    # NOTE: we use 'class_uuid' since StamCSV was initially only meant for classes, or
    # looks like that at least, and itsystems are not classes.. its a bit misleading, but it works.
    @property
    def class_uuid(self) -> uuid.UUID:
        return self.it_system_uuid


class Stillingsbetegnelse(StamCSV):
    job_function_uuid: uuid.UUID = Field(alias="StillingBetUUID")
    job_function: str = Field(alias="Stillingsbetegnelse")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Stillingsbetegnelse.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_job_function"

    @property
    def bvn(self) -> str:
        return self.job_function

    @property
    def title(self) -> str:
        return self.job_function

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.job_function_uuid


class BVNStillingsbetegnelse(StamCSV):
    job_function_uuid: uuid.UUID = Field(alias="BVNStillingBetUUID")
    job_function: str = Field(alias="BVNStillingsbetegnelse")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_BVN_Stillingsbetegnelse.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_job_function_bvn"

    @property
    def bvn(self) -> str:
        return self.job_function

    @property
    def title(self) -> str:
        return self.job_function

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.job_function_uuid


class Lederansvar(StamCSV):
    responsibility_uuid: uuid.UUID = Field(alias="LederansvarUUID")
    responsibility: str = Field(alias="Lederansvar")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederansvar.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "responsibility"

    @property
    def bvn(self) -> str:
        return self.responsibility

    @property
    def title(self) -> str:
        return self.responsibility

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.responsibility_uuid


class Lederniveau(StamCSV):
    level_uuid: uuid.UUID = Field(alias="LederniveauUUID")
    level: str = Field(alias="Lederniveau")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederniveau.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_level"

    @property
    def bvn(self) -> str:
        return self.level

    @property
    def title(self) -> str:
        return self.level

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.level_uuid


class Ledertype(StamCSV):
    type_uuid: uuid.UUID = Field(alias="LedertypeUUID")
    type: str = Field(alias="Ledertype")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Ledertype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_type"

    @property
    def bvn(self) -> str:
        return self.type

    @property
    def title(self) -> str:
        return self.type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.type_uuid


class Tilknytningsrolle(StamCSV):
    association_type_uuid: uuid.UUID = Field(alias="TilknytningsrolleUUID")
    role: str = Field(alias="Tilknytningsrolle")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Tilknytningsrolle.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "association_type"

    @property
    def bvn(self) -> str:
        return self.role

    @property
    def title(self) -> str:
        return self.role

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.association_type_uuid


StamCSVType = Union[
    Type[Engagementstype],
    Type[Enhedstype],
    Type[ITSystem],
    Type[Stillingsbetegnelse],
    Type[BVNStillingsbetegnelse],
    Type[Lederansvar],
    Type[Lederniveau],
    Type[Ledertype],
    Type[Tilknytningsrolle],
]


class StamImporter:
    """Import all classes and IT systems from CSV files.
    Imports based on predetermined UUIDs, so these functions should be idempotent.
    Only imports if file has been modified after `last_import`.
    """

    def __init__(self, last_import: datetime):
        super().__init__()
        self.last_import = last_import

    async def run(self):
        logger.info("Starting STAM import")
        tasks = [
            self.handle_engagementstype(),
            self.handle_enhedstype(),
            self.handle_itsystem(),
            self.handle_stillingsbetegnelse(),
            self.handle_bvn_stillingsbetegnelse(),
            self.handle_lederansvar(),
            self.handle_lederniveau(),
            self.handle_ledertype(),
            self.handle_tilknytningsrolle(),
        ]
        await asyncio.gather(*tasks)
        logger.info("STAM import done")

    async def handle_engagementstype(self):
        """
        Process the external 'engagement type' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Engagementstype)
        return await self._create_classes_from_csv(Engagementstype, rows)

    async def handle_enhedstype(self):
        """
        Process the external 'unit type' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Enhedstype)
        return await self._create_classes_from_csv(Enhedstype, rows)

    async def handle_itsystem(self):
        """
        Process the external 'it system' csv file and create, update and terminate items
        if file is newer than last import.

        OBS: ITSystem are different than the other STAM data fields. since they are not
        normal "classes", which is why we dont use the same logic as ex "handle_stillingsbetegnelse"
        """
        settings = config.get_config()
        gql_client = config.setup_gql_client(settings=settings)

        # Load CSV data
        csv_rows = self._load_csv_if_newer(ITSystem)
        if csv_rows is None:
            return
        csv_row_uuids = set(csv_row.it_system_uuid for csv_row in csv_rows)

        # Get all itsystems already in os2mo
        os2mo_it_systems = gql_get_itsystems(gql_client)
        os2mo_it_system_uuids = [uuid.UUID(its["uuid"]) for its in os2mo_it_systems]

        # Compare CSV with itsystems in os2mo
        new_items = [
            csv_row
            for csv_row in csv_rows
            if csv_row.it_system_uuid not in os2mo_it_system_uuids
        ]
        changed_items = list(
            filter(
                lambda csv_row: any(  # type: ignore
                    str(csv_row.class_uuid) == itsystem["uuid"]  # type: ignore
                    and csv_row.title != itsystem["name"]  # type: ignore
                    for itsystem in os2mo_it_systems
                ),
                csv_rows,
            )
        )
        terminted_item_uuids = [
            itsystem_uuid
            for itsystem_uuid in os2mo_it_system_uuids
            if itsystem_uuid not in csv_row_uuids
        ]

        # Run create, update and terminate on found items
        for new_item in new_items:
            gql_create_it_system(gql_client, new_item)

        for changed_item in changed_items:
            gql_update_it_system(gql_client, changed_item)

        for terminated_item_uuid in terminted_item_uuids:
            gql_terminate_it_system(gql_client, terminated_item_uuid)

    async def handle_stillingsbetegnelse(self):
        """
        Process the external 'job_function' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Stillingsbetegnelse)
        return await self._create_classes_from_csv(Stillingsbetegnelse, rows)

    async def handle_bvn_stillingsbetegnelse(self):
        """
        Process the external 'BVNStillingsbetegnelse' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(BVNStillingsbetegnelse)
        return await self._create_classes_from_csv(BVNStillingsbetegnelse, rows)

    async def handle_lederansvar(self):
        """
        Process the external 'Lederansvar' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Lederansvar)
        return await self._create_classes_from_csv(Lederansvar, rows)

    async def handle_lederniveau(self):
        """
        Process the external 'Lederniveau' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Lederniveau)
        return await self._create_classes_from_csv(Lederniveau, rows)

    async def handle_ledertype(self):
        """
        Process the external 'Ledertype' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Ledertype)
        return await self._create_classes_from_csv(Ledertype, rows)

    async def handle_tilknytningsrolle(self):
        """
        Process the external 'Tilknytningsrolle' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Tilknytningsrolle)
        return await self._create_classes_from_csv(Tilknytningsrolle, rows)

    def _load_csv_if_newer(
        self, csv_class: StamCSVType
    ) -> Union[List[StamCSVType], None]:
        filename = csv_class.get_filename()
        fileset = los_files.get_fileset_implementation()
        try:
            modified_datetime = fileset.get_modified_datetime(filename)
        except ValueError:
            # Raised by `fileset.get_modified_datetime` if file could not be
            # found.
            return None
        else:
            if modified_datetime <= self.last_import:
                return None
            return los_files.read_csv(filename, csv_class)

    async def _create_classes_from_csv(
        self, csv_class: StamCSVType, rows: Optional[List[StamCSVType]] = None
    ) -> None:
        if rows is None:
            return

        settings = config.get_config()
        mox_helper = await create_mox_helper(settings.mox_base)

        facet_uuid = await self._get_or_create_facet(csv_class, mox_helper)

        # Find class UUIDs currently in facet
        gql_client = config.setup_gql_client(settings=settings)
        classes = gql_get_classes(gql_client, str(facet_uuid))
        facet_class_uuids = [uuid.UUID(c["uuid"]) for c in classes]

        # Find rows to insert (classes in CSV but not yet in LoRa)
        rows_class_uuids = set(row.class_uuid for row in rows)
        rows_to_insert = [
            row for row in rows if row.class_uuid not in facet_class_uuids
        ]

        # Create classes in LoRa
        for row in rows_to_insert:
            klasse = mox_payloads.lora_klasse(
                bvn=row.bvn,
                title=row.title,
                facet_uuid=str(facet_uuid),
                org_uuid=str(uuids.ORG_UUID),
                scope="TEXT",
            )
            await mox_helper.insert_klassifikation_klasse(klasse, str(row.class_uuid))
            logger.info("Created LoRa class %r", row.class_uuid)

        # Find classes with updated data
        classes_with_changes = list(
            filter(
                lambda csv_row: any(  # type: ignore
                    str(csv_row.class_uuid) == lora_class["uuid"]  # type: ignore
                    and csv_row.title != lora_class["full_name"]  # type: ignore
                    for lora_class in classes
                ),
                rows,
            )
        )

        for changed_class in classes_with_changes:
            gql_create_class(gql_client, changed_class, str(uuids.ORG_UUID), facet_uuid)  # type: ignore

        # Find classes to terminate (classes in LoRa but no longer in CSV file)
        class_uuids_to_terminate = [
            class_uuid
            for class_uuid in facet_class_uuids
            if class_uuid not in rows_class_uuids
        ]

        # Terminate classes in LoRa by setting them as unpublished
        unpublish_payload = self._get_lora_unpublish_payload()
        for class_uuid in class_uuids_to_terminate:
            try:
                await mox_helper._update(
                    "klassifikation",
                    "klasse",
                    str(class_uuid),
                    unpublish_payload,
                )
            except KeyError:  # this class is already unpublished
                logger.debug("LoRa class %r was already unpublished", class_uuid)
            except ClientResponseError as e:
                # `MoxHelper._update` can now raise a ClientResponseError, rather than a
                # KeyError to indicate an HTTP error.
                if e.status == HttpBadRequest.code:
                    # Based on empirical observation we assume that an HTTP 400 is
                    # caused by attempting to unpublish the same class more than once.
                    logger.debug("LoRa class %r was already unpublished", class_uuid)
                else:
                    raise  # re-raise ClientResponseError
            else:
                logger.info("Unpublished LoRa class %r", class_uuid)

    @staticmethod
    def _get_lora_unpublish_payload():
        return {
            "attributter": {
                "klasseegenskaber": [
                    {
                        "brugervendtnoegle": "-",
                        "titel": "-",
                        "virkning": {
                            "from": "1910-01-01 00:00:00+01",
                            "to": "infinity",
                        },
                    }
                ]
            },
            "tilstande": {
                "klassepubliceret": [
                    {
                        "publiceret": "IkkePubliceret",
                        "virkning": {
                            "from": "1910-01-01 00:00:00+01",
                            "to": "infinity",
                        },
                    }
                ]
            },
        }

    @staticmethod
    async def _get_or_create_facet(csv_class: StamCSVType, mox_helper: Any) -> str:
        facet_bvn = csv_class.get_facet_bvn()

        try:
            # Get facet by its BVN
            facet_uuid = await mox_helper.read_element_klassifikation_facet(
                bvn=facet_bvn
            )
        except ElementNotFound:
            # Facet does not yet exist, create it based on the BVN
            logger.info("LoRa facet %r not found, creating ...", facet_bvn)
            facet_uuid = uuids.uuid_gen(facet_bvn)
            org_uuid = (await mox_helper.read_all_organisation_organisation())[0]
            kls_uuid = (await mox_helper.read_all_klassifikation_klassifikation())[0]
            facet = Facet(
                user_key=facet_bvn,
                uuid=facet_uuid,
                organisation_uuid=org_uuid,
                klassifikation_uuid=kls_uuid,
            )
            await mox_helper.insert_klassifikation_facet(facet.build(), facet_uuid)

        return facet_uuid


def gql_get_classes(gql_client: GraphQLClient, facet_uuid: str) -> List[dict]:
    graphql_query = gql(
        """
        query GetFacetClasses($uuid: [UUID!]) {
        classes(facets: $uuid) {
                uuid
                full_name
                published
            }
        }
        """
    )

    response = gql_client.execute(graphql_query, variable_values={"uuid": [facet_uuid]})
    return response["classes"]


def gql_create_class(
    gql_client: GraphQLClient,
    stamp_csv_row: StamCSV,
    org_uuid: str,
    facet_uuid: str,
):
    graphql_query = gql(
        """
        mutation CreateClass($input: ClassCreateInput!) {
            class_create(input: $input) {
                uuid
            }
        }
        """
    )

    response = gql_client.execute(
        graphql_query,
        variable_values={
            "input": {
                "uuid": str(stamp_csv_row.class_uuid),
                "org_uuid": org_uuid,
                "facet_uuid": facet_uuid,
                "name": stamp_csv_row.title,
                "user_key": stamp_csv_row.bvn,
                "scope": "TEXT",
            }
        },
    )
    return response["class_create"]


def gql_get_itsystems(gql_client: GraphQLClient) -> List[dict]:
    graphql_query = gql(
        """
        query GetITSystems {
            itsystems {
                uuid
                name
                user_key
            }
        }
        """
    )

    response = gql_client.execute(graphql_query)
    return response["itsystems"]


def gql_create_it_system(
    gql_client: GraphQLClient,
    stamp_csv_row: StamCSV,
    from_date: date | None = None,
    to_date: date | None = None,
):
    graphql_query = gql(
        """
        mutation CreateITSystem($input: ITSystemCreateInput!) {
            itsystem_create(input: $input) {
                uuid
            }
        }
        """
    )

    response = gql_client.execute(
        graphql_query,
        variable_values={
            "input": {
                "uuid": str(stamp_csv_row.class_uuid),
                "name": stamp_csv_row.title,
                "user_key": stamp_csv_row.bvn,
                "from": from_date.isoformat() if from_date else None,
                "to": to_date.isoformat() if to_date else None,
            }
        },
    )

    return response["itsystem_create"]


def gql_update_it_system(
    gql_client: GraphQLClient,
    stamp_csv_row: StamCSV,
    from_date: date | None = None,
    to_date: date | None = None,
):
    graphql_query = gql(
        """
        mutation UpdateITSystem($uuid: UUID!, $input: ITSystemCreateInput!) {
            itsystem_update(uuid: $uuid, input: $input) {
                uuid
            }
        }
        """
    )

    response = gql_client.execute(
        graphql_query,
        variable_values={
            "uuid": str(stamp_csv_row.class_uuid),
            "input": {
                "name": stamp_csv_row.title,
                "user_key": stamp_csv_row.bvn,
                "from": from_date.isoformat() if from_date else None,
                "to": to_date.isoformat() if to_date else None,
            },
        },
    )

    return response["itsystem_update"]


def gql_terminate_it_system(gql_client: GraphQLClient, it_system_uuid: uuid.UUID):
    # TODO: Implement this when its possible to terminate itsystem bi-temporally
    logger.warning(
        f"IGNORED terminate of itsystem, {str(it_system_uuid)}, due to missing implementation of itsystem bitemporality"
    )
    return



import json
import xml
from collections import defaultdict
from operator import itemgetter

import click
import xmltodict
from asciitree import LeftAligned


def dict_map(dicty, key_fun=None, value_fun=None, tuple_fun=None):
    """Map the dict values.

    Example:
        >>> dts = lambda dicty: sorted(dicty.items())
        >>> dicty = {1: 1, 2: 2, 3: 3}
        >>> dts(dict_map(dicty, value_fun=lambda value: value ** 2))
        [(1, 1), (2, 4), (3, 9)]
        >>> dts(dict_map(dicty, key_fun=lambda key: key + 2))
        [(3, 1), (4, 2), (5, 3)]
        >>> dts(dict_map(dicty, tuple_fun=lambda key, value: (key + 2, value ** 2)))
        [(3, 1), (4, 4), (5, 9)]

    Returns:
        dict: A dict where func has been applied to every value.
    """
    key_fun = key_fun or (lambda key: key)
    value_fun = value_fun or (lambda value: value)
    tuple_fun = tuple_fun or (lambda key, value: (key, value))
    return dict(
        tuple_fun(key_fun(key), value_fun(value)) for key, value in dicty.items()
    )


def build_children_map(parent_map):
    """
    Build children map from parent map.

    Roots are assumed to have parent of :code:`None`

    Example:
        >>> parent_map = {1: None, 2: 1, 3: 2, 4: 1, 5: 2}
        >>> expected = {None: [1], 1: [2, 4], 2: [3, 5]}
        >>> build_children_map(parent_map) == expected
        True

    Args:
        parent_map: dict from id to parent id

    Returns:
        dict: A dictionary from parent-id to list of children
    """
    children_map = defaultdict(list)
    for idd, parent in parent_map.items():
        children_map[parent].append(idd)
    children_map = dict(children_map)
    return children_map


def recursive_build_tree(children_map, entry=None):
    """Recursively build tree structure from children map.

    Roots are assumed to be parents of :code:`None`.

    Example:
        >>> children_map = {None: [1], 1: [2, 4], 2: [3, 5]}
        >>> expected = {1: {2: {3: {}, 5: {}}, 4: {}}}
        >>> recursive_build_tree(children_map) == expected
        True

    Args:
        children_map: dict from parent-id to list of children
        entry: key from children_map to build subtree for

    Returns:
        dict: Tree structure
    """
    output = {}
    for root in children_map.get(entry, []):
        output[root] = {}
        for entry in children_map.get(root, []):
            output[root][entry] = recursive_build_tree(children_map, entry)
    return output


def build_tree(parent_map):
    """
    >>> parent_map = {1: None, 2: 1, 3: 2, 4: 1, 5: 2}
    >>> expected = {1: {2: {3: {}, 5: {}}, 4: {}}}
    >>> build_tree(parent_map) == expected
    True
    """
    children_map = build_children_map(parent_map)
    return recursive_build_tree(children_map)


def dict_from_xmlfile(filename):
    def read_file(filename):
        with open(filename, "r") as input_file:
            data = input_file.read()
            return data

    def read_json_file(filename):
        return json.loads(read_file(filename))

    def write_json_file(filename, data):
        with open(filename, "w") as output_file:
            json.dump(data, output_file)

    # Read cache, if any
    cache_filename = filename + ".json"
    try:
        return read_json_file(cache_filename)
    # No cache, no problem, read xml source and produce cache
    except FileNotFoundError:
        xml_data = read_file(filename)
        data = xmltodict.parse(xml_data)
        write_json_file(cache_filename, data)
        return data


def print_tree(tree, print_map):
    tr = LeftAligned()
    tr.traverse.get_text = lambda node: print_map[node[0]]
    print(tr(tree))


@click.command()
@click.argument("filename", type=click.Path(exists=True))
def output_tree(filename):
    """Load organisation units from [FILENAME] and print org-tree."""
    # Load data and process it
    try:
        data = dict_from_xmlfile(filename)
    except xml.parsers.expat.ExpatError:
        raise click.ClickException("Provided file not in XML format")
    try:
        org_units = data["kmd"]["orgUnit"]
        parent_map = dict(map(itemgetter("@id", "parentOrgUnit"), org_units))
        print_map = dict_map(
            dict(map(itemgetter("@id", "longName"), org_units)),
            tuple_fun=lambda key, value: (key, value + " (" + key + ")"),
        )
    except KeyError:
        raise click.ClickException("Unexpected XML")

    tree = build_tree(parent_map)
    print_tree(tree, print_map)


if __name__ == "__main__":
    output_tree()



"""ad_fix_enddate.py

This program takes engagement end dates from MO and writes them to the corresponding AD
user.

The program has evolved from a "one-time fix script" to be responsible for all updates
of end dates in AD. In other words, it has "taken over" writing end dates from
"mo_to_ad_sync.py" (which is unable to "see" past end dates, as it only considers
engagements in the present or future, and thus cannot process terminated engagements.)

The program currently supports two modes of operation:
    a) Writing a single MO end date to a single AD field on each AD user.
    b) Writing two MO end dates to two AD fields on each AD user.

In the first mode, the end date to write is the latest end date of all engagements found
for a given MO user.

In the second mode, the engagement end dates are split into two:
    - the "current" end date is the latest end date occurring in either the past or
      present.
    - the "future" end date is the first end date occurring in the future.
"""
import datetime
import logging
from dataclasses import dataclass
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import Self

import click
import httpx
import sentry_sdk
from dateutil import tz
from fastapi.encoders import jsonable_encoder
from gql import gql
from more_itertools import one
from more_itertools import partition
from ra_utils.job_settings import JobSettings
from ra_utils.load_settings import load_setting
from ra_utils.tqdm_wrapper import tqdm
from raclients.graph.client import GraphQLClient
from raclients.graph.client import SyncClientSession
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_delay
from tenacity import wait_fixed

from integrations.ad_integration.ad_common import AD
from integrations.ad_integration.ad_reader import ADParameterReader
from integrations.ad_integration.ad_template_engine import render_update_by_mo_uuid_cmd


logger = logging.getLogger(__name__)

DEFAULT_TIMEZONE = tz.gettz("Europe/Copenhagen")


def in_default_tz(val: datetime.datetime) -> datetime.datetime:
    return val.astimezone(DEFAULT_TIMEZONE)


class _SymbolicConstant:
    # Helper class which is always equal to itself

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return True
        return super().__eq__(other)


class Unset(_SymbolicConstant):
    """Represents an end date that should *not* be updated in AD.
    Typically because the MO user has no engagements at all, or because the MO user has
    no engagements in the future, or the past/present.

    When encountering an `Unset` end date, the program should not update the matching
    AD field.
    """


class Invalid(_SymbolicConstant):
    """Represents an invalid AD end date, usually due to malformed or unexpected data,
    that cannot be parsed into a `datetime.datetime` object.
    """


class PositiveInfinity(_SymbolicConstant):
    """Represents the upper bound of a validity period if MO returns `None`, i.e. the
    validity period ends at "infinity."
    """

    def as_datetime(self) -> datetime.datetime:
        return in_default_tz(datetime.datetime.fromisoformat("9999-12-31"))


class NegativeInfinity(_SymbolicConstant):
    """Represents the lower bound of a validity period if MO returns `None`, i.e. the
    validity period begins at "-infinity."
    """

    def as_datetime(self) -> datetime.datetime:
        return in_default_tz(datetime.datetime.fromisoformat("1930-01-01"))


class AdFixEndDateSettings(JobSettings):
    # Currently only used for calling `.start_logging_based_on_settings` in `cli`

    class Config:
        settings_json_prefix = "integrations.ad.write"


@dataclass
class ADDate:
    """Represents a date found on an AD user"""

    field_name: str
    field_value: str | None | Invalid

    @property
    def normalized_value(self) -> datetime.datetime | None | Invalid:
        if self.field_value == Invalid():
            return Invalid()
        try:
            return in_default_tz(datetime.datetime.fromisoformat(self.field_value))  # type: ignore
        except (TypeError, ValueError):
            logger.debug(
                "%s: cannot parse %r as ISO datetime",
                self.field_name,
                self.field_value,
            )
            return Invalid()


@dataclass
class ADText:
    """Represents a text field found on an AD user"""

    field_name: str
    field_value: str | None | Invalid


@dataclass
class ADUser:
    """Represents an AD user and its associated data"""

    mo_uuid: str
    end_date: ADDate
    end_date_future: ADDate | None = None
    start_date_future: ADDate | None = None
    org_unit_path: ADText | None = None


class ADUserSource:
    """Fetch AD data for one or all AD users.
    Iterating over an `ADUserSource` instance produces one or more `ADUser` instances.
    """

    def __init__(
        self,
        uuid_field: str,
        enddate_field: str,
        enddate_field_future: str | None,
        startdate_field_future: str | None,
        org_unit_path_field_future: str | None,
        settings: dict | None = None,
    ):
        self._uuid_field = uuid_field
        self._enddate_field = enddate_field
        self._enddate_field_future = enddate_field_future
        self._startdate_field_future = startdate_field_future
        self._org_unit_path_field_future = org_unit_path_field_future

        self._reader = ADParameterReader(all_settings=settings)

        # This holds the "raw" result of calling either `ADParameterReader.read_it_all`
        # (in `of_all_users`), or `ADParameterReader.read_user` (in `of_one_user`.)
        self._ad_users: list[dict] = []

    def __iter__(self) -> Iterator[ADUser]:
        """Yield an `ADUser` instance for each AD user in `self._ad_users`.
        Iterating over an `ADUserSource` only makes sense after calling either
        `of_all_users` or `of_one_user`.
        """
        for ad_user in self._ad_users:
            if self._uuid_field not in ad_user:
                click.echo(
                    f"User with {ad_user['ObjectGuid']=} does not have an "
                    f"{self._uuid_field} field, and will be skipped"
                )
                continue

            yield ADUser(
                ad_user[self._uuid_field],
                self._get_ad_date(ad_user, self._enddate_field),  # type: ignore
                self._get_ad_date(ad_user, self._enddate_field_future),
                self._get_ad_date(ad_user, self._startdate_field_future),
                self._get_ad_text(ad_user, self._org_unit_path_field_future),
            )

    def of_all_users(self) -> Self:
        """Return `ADUserSource` for all AD users"""
        self._ad_users = self._reader.read_it_all()
        return self

    def of_one_user(self, username: str) -> Self:
        """Return `ADUserSource` for a single AD user given by its AD `username`"""
        self._ad_users = [self._reader.read_user(user=username)]
        return self

    def _get_ad_date(self, ad_user: dict, field_name: str | None) -> ADDate | None:
        if field_name is not None:
            return ADDate(field_name, self._get_case_insensitive(ad_user, field_name))
        return None

    def _get_ad_text(self, ad_user: dict, field_name: str | None) -> ADText | None:
        if field_name is not None:
            return ADText(field_name, self._get_case_insensitive(ad_user, field_name))
        return None

    def _get_case_insensitive(
        self, ad_user: dict, field_name: str
    ) -> str | Invalid | None:
        # There may be case inconsistencies in AD field names between the name we ask
        # for, and the name that is actually returned. E.g. asking for a field name in
        # all lowercase may return the same field but with a mixed-case name.
        # This method handles this by treating both field names as lowercase.
        #
        # If the field name is not present in the given AD user, an `Invalid` instance
        # is returned.
        try:
            return one(
                (v for k, v in ad_user.items() if k.lower() == field_name.lower())
            )
        except ValueError:
            return Invalid()


@dataclass
class MOSimpleEngagement:
    """Represents the MO data to consider in the "simple" case, where we are only
    updating a single end date in AD.

    This class only supports comparisons on the `end_date` field.
    """

    ad_user: ADUser
    end_date: datetime.datetime | PositiveInfinity | Unset

    @property
    def changes(self) -> dict[str, str]:
        """Return a dictionary where the keys are AD field names and the values are
        strings, e.g. the contents are suitable for passing to
        `ChangeListExecutor.get_update_cmd`.
        """
        result: dict[str, str] = {}
        new_end_date: datetime.datetime | None
        new_end_date = self._compare_end_date(self.ad_user.end_date, self.end_date)
        result = self._add_changed_date(self.ad_user.end_date, new_end_date, result)
        return result

    def _compare_end_date(
        self,
        ad_date: ADDate | None,
        mo_date: datetime.datetime | PositiveInfinity | Unset | None,
    ) -> datetime.datetime | None:
        # Convert `None` to "max date"
        if mo_date is None:
            mo_date = PositiveInfinity().as_datetime()
        return self._compare_date(ad_date, mo_date)

    def _compare_date(
        self,
        ad_date: ADDate | None,
        mo_value: datetime.datetime | PositiveInfinity | NegativeInfinity | Unset,
    ) -> datetime.datetime | None:
        # Parse AD date to `datetime.datetime` if available, else use None as its value
        ad_value = ad_date.normalized_value if ad_date else None

        # Convert `PositiveInfinity` and `NegativeInfinity` to their respective
        # `datetime.datetime` values.
        if mo_value in (PositiveInfinity(), NegativeInfinity()):
            mo_value = mo_value.as_datetime()  # type: ignore

        if mo_value == Unset():
            # If MO value is `Unset`, leave it out of the set of changes
            logger.info(
                "MO user %r: skipping %r update as MO value is unset",
                self.ad_user.mo_uuid,
                ad_date.field_name if ad_date is not None else "",
            )
            return None
        elif ad_value != mo_value:
            # If MO and AD values differ, include in the list of changes
            return mo_value  # type: ignore
        else:
            # MO and AD values must be equal
            logger.debug(
                "MO user %r: normalized MO and AD values are identical in field %r "
                "(values in MO: %r, AD: %r)",
                self.ad_user.mo_uuid,
                ad_date.field_name if ad_date is not None else None,
                mo_value.strftime("%Y-%m-%d"),  # type: ignore
                ad_date.field_value if ad_date is not None else None,
            )
            return None

    def _add_changed_date(
        self,
        ad_date: ADDate | None,
        new_value: datetime.datetime | None,
        changes: dict,
    ) -> dict:
        if ad_date is not None and new_value is not None:
            changes[ad_date.field_name] = new_value.strftime("%Y-%m-%d")
        return changes


@dataclass
class MOSplitEngagement(MOSimpleEngagement):
    """Represents the MO data to consider in the complex case, where we look at both the
    current/past engagement, as well as a possible future engagement.

    This class extends `MOSimpleEngagement` to support comparisons on the additional
    fields `end_date_future`, `start_date_future`, and `org_unit_path_future`.
    """

    end_date_future: datetime.datetime | PositiveInfinity | Unset
    start_date_future: datetime.datetime | NegativeInfinity | Unset
    org_unit_path_future: str | Unset

    @property
    def changes(self) -> dict[str, str]:
        # Compare end_date
        result: dict[str, str] = super().changes or {}

        # Compare end_date_future
        new_end_date_future: datetime.datetime | None
        new_end_date_future = self._compare_end_date(
            self.ad_user.end_date_future,
            self.end_date_future,
        )
        result = self._add_changed_date(
            self.ad_user.end_date_future,
            new_end_date_future,
            result,
        )

        # Compare start_date_future
        new_start_date_future: datetime.datetime | None
        new_start_date_future = self._compare_start_date(
            self.ad_user.start_date_future,
            self.start_date_future,
        )
        result = self._add_changed_date(
            self.ad_user.start_date_future,
            new_start_date_future,
            result,
        )

        # Compare org_unit_path_future
        new_org_unit_path_future: str | None
        new_org_unit_path_future = self._compare_org_unit_path(
            self.ad_user.org_unit_path,
            self.org_unit_path_future,
        )
        result = self._add_changed_text(
            self.ad_user.org_unit_path,
            new_org_unit_path_future,
            result,
        )

        return result

    def _compare_start_date(
        self,
        ad_date: ADDate | None,
        mo_date: datetime.datetime | NegativeInfinity | Unset | None,
    ) -> datetime.datetime | None:
        # Convert `None` to "min date"
        if mo_date is None:
            mo_date = NegativeInfinity().as_datetime()
        return self._compare_date(ad_date, mo_date)

    def _compare_org_unit_path(self, ad_text: ADText | None, mo_value) -> str | None:
        ad_value: str | None | Invalid = (
            ad_text.field_value if ad_text is not None else None
        )
        if mo_value == Unset():
            logger.info(
                "MO user %r: skipping %r update as MO value is unset",
                self.ad_user.mo_uuid,
                ad_text.field_name if ad_text is not None else "",
            )
            return None
        elif ad_value != mo_value:
            # If MO and AD values differ, include in the list of changes
            return mo_value
        else:
            # MO and AD values must be equal
            logger.debug(
                "MO user %r: normalized MO and AD values are identical in field %r "
                "(values in MO: %r, AD: %r)",
                self.ad_user.mo_uuid,
                ad_text.field_name if ad_text is not None else None,
                mo_value,
                ad_text.field_value if ad_text is not None else None,
            )
            return None

    def _add_changed_text(
        self,
        ad_text: ADText | None,
        new_value: str | None,
        changes: dict,
    ) -> dict:
        if ad_text is not None and new_value is not None:
            changes[ad_text.field_name] = new_value
        return changes


@dataclass
class _ParsedEngagement:
    # Only used internally by `MOEngagementSource`

    from_dt: datetime.datetime | NegativeInfinity
    to_dt: datetime.datetime | PositiveInfinity
    org_unit: list[dict]

    def get_org_unit_path(self, sep: str = "\\") -> str:
        if len(self.org_unit) > 1:
            logger.warning(
                "more than one `org_unit` found in %r, taking first",
                self.org_unit,
            )
        path = self.org_unit[0]
        return sep.join(
            elem["name"] for elem in (path["ancestors_validity"][::-1] + [path])
        )


class MOEngagementSource:
    """Fetch engagement data from MO for a given MO user UUID, and return either:
    - a `MOSimpleEngagement` (via `get_simple_engagement`), or
    - a `MOSplitEngagement` (via `get_split_engagement`.)
    """

    def __init__(
        self,
        graphql_session: SyncClientSession,
        split: bool = False,
    ):
        # self._enddate_field = enddate_field
        # self._enddate_field_future = enddate_field_future
        self.split = split
        self._graphql_session: SyncClientSession = graphql_session

    def __getitem__(self, ad_user: ADUser):
        if self.split:
            return self.get_split_engagement(ad_user)
        else:
            return self.get_simple_engagement(ad_user)

    @retry(
        wait=wait_fixed(5),
        reraise=True,
        stop=stop_after_delay(10 * 60),
        retry=retry_if_exception_type(httpx.HTTPError),
    )
    def _get_employee_engagements(self, uuid: str) -> list[dict]:
        """Fetch all MO engagement validity periods and org unit paths for a given MO
        user UUID."""

        query = gql(
            """
            query Get_mo_engagements($employees: [UUID!]) {
                engagements(employees: $employees, from_date: null, to_date: null) {
                    objects {
                        validity {
                            from
                            to
                        }
                        org_unit {
                            name
                            ancestors_validity {
                                name
                            }
                        }
                    }
                }
            }
            """
        )
        result = self._graphql_session.execute(
            query,
            variable_values=jsonable_encoder({"employees": uuid}),
        )
        if not result["engagements"]:
            raise KeyError("User not found in mo")
        return result["engagements"]

    def get_simple_engagement(self, ad_user: ADUser) -> MOSimpleEngagement:
        """Return a `MOSimpleEngagement` for a given MO user UUID"""

        try:
            engagements = self._get_employee_engagements(ad_user.mo_uuid)
        except KeyError:
            return MOSimpleEngagement(ad_user, Unset())

        parsed_engagements = self._parse_engagements(engagements)
        if parsed_engagements:
            end_date: datetime.datetime | PositiveInfinity = max(
                (parsed_engagement.to_dt for parsed_engagement in parsed_engagements),
                key=lambda end_datetime: self._fold_in_utc(
                    end_datetime, datetime.datetime.max
                ),
            )
            return MOSimpleEngagement(ad_user, end_date)
        else:
            return MOSimpleEngagement(ad_user, Unset())

    def get_split_engagement(self, ad_user: ADUser) -> MOSplitEngagement:
        """Return a `MOSplitEngagement` for a given MO user UUID"""

        def _split(parsed_engagement: _ParsedEngagement):
            """Split the set of engagement validities into two partitions:
            - one partition of past and present engagements.
            - one partition of future engagements.
            The split is determined by looking at the "from" part of the validity.
            """
            local_date = in_default_tz(datetime.datetime.now()).date()
            # Take either engagement start date, or "negative infinity" if blank
            start_date = self._fold_in_utc(
                parsed_engagement.from_dt, datetime.datetime.min
            ).date()
            return start_date > local_date

        try:
            engagements = self._get_employee_engagements(ad_user.mo_uuid)
        except KeyError:
            # If MO user has no engagements, we should not update neither the current
            # nor the future end date.
            return MOSplitEngagement(ad_user, Unset(), Unset(), Unset(), Unset())

        parsed_engagements: list[_ParsedEngagement] = self._parse_engagements(
            engagements
        )

        # Split by whether the engagement *starts* in the future, or not
        current: Iterator[_ParsedEngagement]
        future: Iterator[_ParsedEngagement]
        current, future = partition(_split, parsed_engagements)

        # Sort by validity end - the validity with the latest end appears first in list
        current_list: list[_ParsedEngagement] = sorted(
            current,
            key=lambda parsed_engagement: self._fold_in_utc(
                parsed_engagement.to_dt,
                datetime.datetime.max,
            ),
            reverse=True,
        )

        # Sort by validity start - the validity with the earliest start appears first in
        # list.
        future_list: list[_ParsedEngagement] = sorted(
            future,
            key=lambda parsed_engagement: self._fold_in_utc(
                parsed_engagement.from_dt,
                datetime.datetime.min,
            ),
            reverse=False,
        )

        # Check if either `current` or `future` (or both) is a 0-length list, and return
        # results accordingly.
        if current_list and future_list:
            # Return end date of the latest current validity and the earliest future
            # validity.
            return MOSplitEngagement(
                ad_user,
                current_list[0].to_dt,  # end_date
                future_list[0].to_dt,  # end_date_future
                future_list[0].from_dt,  # start_date_future
                future_list[0].get_org_unit_path(),  # org_unit_path_future
            )
        elif current_list and not future_list:
            # Return only end date of the latest current validity
            return MOSplitEngagement(
                ad_user,
                current_list[0].to_dt,  # end_date
                Unset(),  # end_date_future
                Unset(),  # start_date_future
                Unset(),  # org_unit_path_future
            )
        elif not current_list and future_list:
            # Return only end date of the earliest future validity
            return MOSplitEngagement(
                ad_user,
                Unset(),  # end_date
                future_list[0].to_dt,  # end_date_future
                future_list[0].from_dt,  # start_date_future
                future_list[0].get_org_unit_path(),  # org_unit_path_future
            )

        # No engagements at all
        return MOSplitEngagement(ad_user, Unset(), Unset(), Unset(), Unset())

    def _fold_in_utc(
        self,
        val: datetime.datetime | None | PositiveInfinity | NegativeInfinity,
        default: datetime.datetime,
    ) -> datetime.datetime:
        # Fold in UTC as `datetime.datetime.max` and `datetime.datetime.min` cannot be
        # converted into other timezones without surprising issues.
        return (
            default
            if val in (None, PositiveInfinity(), NegativeInfinity())
            else val.astimezone(datetime.timezone.utc).replace(tzinfo=None)  # type: ignore
        )

    def _parse_engagements(self, engagements) -> list[_ParsedEngagement]:
        """Convert validities from string values to 2-tuple of `datetime.datetime`
        objects.
        """

        def _from_iso_or_none(val: str, none: PositiveInfinity | NegativeInfinity):
            if val:
                maybe_naive_dt = datetime.datetime.fromisoformat(val)
                if maybe_naive_dt.tzinfo:
                    return maybe_naive_dt
                else:
                    return in_default_tz(maybe_naive_dt)
            # If val is None or '', convert to the `none` value provided by the caller,
            # either `PositiveInfinity` or `NegativeInfinity`.
            return none

        def _start_from_iso_or_none(val: str) -> datetime.datetime | NegativeInfinity:
            return _from_iso_or_none(val, NegativeInfinity())

        def _end_from_iso_or_none(val: str) -> datetime.datetime | PositiveInfinity:
            return _from_iso_or_none(val, PositiveInfinity())

        def _parse(engagement: dict) -> _ParsedEngagement:
            validity = engagement["validity"]
            org_unit = engagement["org_unit"]

            start = _start_from_iso_or_none(validity["from"])
            end = _end_from_iso_or_none(validity["to"])

            # Sanity check that engagement start is indeed before engagement end
            folded_start = self._fold_in_utc(start, datetime.datetime.min)
            folded_end = self._fold_in_utc(end, datetime.datetime.max)
            assert folded_start <= folded_end

            return _ParsedEngagement(start, end, org_unit)

        parsed_validities = [
            _parse(obj) for engagement in engagements for obj in engagement["objects"]
        ]

        return parsed_validities


class ChangeList:
    """Compares the data in `mo_engagement_source` and `ad_user_source`, and produces a
    list of changes that need to be made in AD.
    """

    def __init__(
        self,
        mo_engagement_source: MOEngagementSource,
        ad_user_source: ADUserSource,
    ):
        self._mo_engagement_source = mo_engagement_source
        self._ad_user_source = ad_user_source

    def get_changes(self) -> Iterator[MOSimpleEngagement | MOSplitEngagement]:
        """For each AD user end date in `self._ad_user_source`, produce the relevant
        MO data (either a `MOSimpleEngagement` or a `MOSplitEngagement`.)
        """
        for ad_user in self._ad_user_source:
            match: MOSimpleEngagement | MOSplitEngagement | None
            match = self._mo_engagement_source[ad_user]
            if match is not None and match.changes:
                yield match


class ChangeListExecutor(AD):
    """Given a list of changes, perform update commands against AD to bring the data of
    the AD users up-to-date with the corresponding MO users.
    """

    def __init__(self, settings=None):
        super().__init__(all_settings=settings)

    def get_update_cmd(
        self,
        uuid_field: str,
        uuid: str,
        **changes,
    ) -> str:
        """Return the relevant Powershell update command to update the given end date
        field in AD with the given value."""
        return self.remove_redundant(
            render_update_by_mo_uuid_cmd(
                self._ps_boiler_plate()["complete"],
                self._ps_boiler_plate()["credentials"],
                uuid_field,
                uuid,
                changes,
            )
        )

    def run(self, cmd) -> dict:
        """Run a PowerShell command against AD"""
        return self._run_ps_script("%s\n%s" % (self._build_user_credential(), cmd))

    def run_all(
        self,
        changes: Iterable[MOSimpleEngagement | MOSplitEngagement],
        uuid_field: str,
        dry: bool = False,
    ) -> list[tuple[str, Any]]:
        """Take the output of `ChangeList.get_changes` and turn it into update
        commands. Run the update commands against AD if `dry` is False, or else just log
        the commands that would have been run.
        """

        changes = tqdm(list(changes))
        num_changes = 0
        retval = []

        for change in changes:
            if change.changes == {}:
                logger.debug("skip unchanged user %r", change.ad_user)
                continue
            cmd = self.get_update_cmd(
                uuid_field,
                change.ad_user.mo_uuid,
                **change.changes,
            )
            logger.info(
                "Updating AD user %r: %r",
                change.ad_user.mo_uuid,
                change.changes,
            )
            logger.debug(cmd)
            if dry:
                retval.append((cmd, "<dry run>"))
            else:
                result = self.run(cmd)
                retval.append((cmd, result))  # type: ignore
                if result != {}:
                    logger.error("AD error response %r", result)
                else:
                    num_changes += 1

        logger.info("%d users end dates corrected", num_changes)
        logger.info("All end dates are fixed")

        return retval


@click.command()
@click.option(
    "--enddate-field",
    default=load_setting("integrations.ad_writer.fixup_enddate_field"),
)
@click.option(
    "--enddate-field-future",
    default=load_setting("integrations.ad_writer.fixup_enddate_field_future", None),
)
@click.option(
    "--startdate-field-future",
    default=load_setting("integrations.ad_writer.fixup_future_startdate_field", None),
)
@click.option(
    "--orgunitpath-field-future",
    default=load_setting("integrations.ad_writer.fixup_future_orgunitpath_field", None),
)
@click.option("--uuid-field", default=load_setting("integrations.ad.write.uuid_field"))
@click.option(
    "--ad-user",
    help="If given, update only one AD user (specified by username)",
)
@click.option("--dry-run", is_flag=True)
@click.option("--mora-base", envvar="MORA_BASE", default="http://mo")
@click.option("--client-id", envvar="CLIENT_ID", default="dipex")
@click.option("--client-secret", envvar="CLIENT_SECRET")
@click.option("--auth-realm", envvar="AUTH_REALM", default="mo")
@click.option("--auth-server", envvar="AUTH_SERVER", default="http://keycloak")
def cli(
    enddate_field,
    enddate_field_future,
    startdate_field_future,
    orgunitpath_field_future,
    uuid_field,
    ad_user,
    dry_run,
    mora_base: str,
    client_id: str,
    client_secret: str,
    auth_realm: str,
    auth_server: str,
):
    """Writes MO engagements end date(s) to one or more fields on the corresponding AD
    user(s).
    """
    pydantic_settings = AdFixEndDateSettings()
    pydantic_settings.start_logging_based_on_settings()

    if pydantic_settings.sentry_dsn:
        sentry_sdk.init(dsn=pydantic_settings.sentry_dsn)

    logger.info(
        f"Command line args:"
        f" end-date-field = {enddate_field},"
        f" end-date-field-future = {enddate_field_future},"
        f" start-date-field-future = {startdate_field_future},"
        f" org-unit-path-field-future = {orgunitpath_field_future},"
        f" uuid-field = {uuid_field},"
        f" dry-run = {dry_run},"
        f" mora-base = {mora_base},"
        f" client-id = {client_id},"
        f" client-secret = not logged,"
        f" auth-realm = {auth_realm},"
        f" auth-server = {auth_server}",
    )

    graphql_client = GraphQLClient(
        url=f"{mora_base}/graphql/v3",
        client_id=client_id,
        client_secret=client_secret,
        auth_realm=auth_realm,
        auth_server=auth_server,
        sync=True,
        httpx_client_kwargs={"timeout": None},
    )

    ad_user_source = ADUserSource(
        uuid_field,
        enddate_field,
        enddate_field_future,
        startdate_field_future,
        orgunitpath_field_future,
    )
    if ad_user:
        ad_user_source = ad_user_source.of_one_user(ad_user)
    else:
        ad_user_source = ad_user_source.of_all_users()

    with graphql_client as session:
        mo_engagement_source = MOEngagementSource(
            session,
            split=True if enddate_field_future else False,
        )
        change_list = ChangeList(mo_engagement_source, ad_user_source)
        executor = ChangeListExecutor()
        executor.run_all(change_list.get_changes(), uuid_field, dry=dry_run)


if __name__ == "__main__":
    cli()



import datetime
import logging
from typing import Iterator
from unittest.mock import Mock
from unittest.mock import patch

import pytest as pytest
from click.testing import CliRunner
from freezegun import freeze_time
from hypothesis import given
from hypothesis import strategies as st

from ..ad_fix_enddate import _ParsedEngagement
from ..ad_fix_enddate import ADDate
from ..ad_fix_enddate import ADText
from ..ad_fix_enddate import ADUser
from ..ad_fix_enddate import ADUserSource
from ..ad_fix_enddate import ChangeList
from ..ad_fix_enddate import ChangeListExecutor
from ..ad_fix_enddate import cli
from ..ad_fix_enddate import DEFAULT_TIMEZONE
from ..ad_fix_enddate import Invalid
from ..ad_fix_enddate import MOEngagementSource
from ..ad_fix_enddate import MOSimpleEngagement
from ..ad_fix_enddate import MOSplitEngagement
from ..ad_fix_enddate import NegativeInfinity
from ..ad_fix_enddate import PositiveInfinity
from ..ad_fix_enddate import Unset
from ..ad_reader import ADParameterReader
from .mocks import AD_UUID_FIELD
from .mocks import MO_UUID
from .mocks import MockADParameterReader
from .mocks import MockADParameterReaderWithMOUUID


ENDDATE_FIELD = "enddate_field"
ENDDATE_FIELD_FUTURE = "enddate_field_future"
ORG_UNIT_PATH_FIELD_FUTURE = "orgunitpath_field_future"
TEST_SEARCH_BASE = "search_base"
TEST_SETTINGS = {
    "primary": {
        "search_base": TEST_SEARCH_BASE,
        "system_user": "username",
        "password": "password",
    },
}
VALID_AD_DATE = "2020-01-01"
VALID_AD_DATE_FUTURE = "2025-01-01"
INVALID_AD_DATE = "<invalid date>"
AD_ORG_UNIT_PATH_EMPTY = ADText(ORG_UNIT_PATH_FIELD_FUTURE, "")
AD_USER = ADUser(MO_UUID, None)  # type: ignore
AD_USER_VALID_DATE = ADUser(MO_UUID, ADDate(ENDDATE_FIELD, VALID_AD_DATE))
AD_USER_VALID_DATE_FUTURE = ADUser(
    MO_UUID,
    ADDate(ENDDATE_FIELD, VALID_AD_DATE),
    ADDate(ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
    None,  # start_date_future
    AD_ORG_UNIT_PATH_EMPTY,
)


class MockADParameterReaderWithMOUUIDAndInvalidDate(MockADParameterReaderWithMOUUID):
    def read_user(self, **kwargs):
        ad_user = super().read_user(**kwargs)
        ad_user[ENDDATE_FIELD] = INVALID_AD_DATE
        return ad_user


class MockADParameterReaderWithMOUUIDAndValidSplitDates(
    MockADParameterReaderWithMOUUID
):
    def read_user(self, **kwargs):
        ad_user = super().read_user(**kwargs)
        ad_user[ENDDATE_FIELD] = VALID_AD_DATE
        ad_user[ENDDATE_FIELD_FUTURE] = VALID_AD_DATE_FUTURE
        return ad_user


class _MockADUserSource(ADUserSource):
    def __init__(self):
        # Explicitly avoid calling `super().__init__(...)` as we don't want to create an
        # `ADParameterReader` instance.
        pass

    def __iter__(self) -> Iterator[ADUser]:
        raise NotImplementedError("must be implemented by subclass")


class _MockADUserSourceNoMatchingADUser(_MockADUserSource):
    def __iter__(self) -> Iterator[ADUser]:
        return iter([])


class _MockADUserSourceMatchingADUser(_MockADUserSource):
    def __iter__(self) -> Iterator[ADUser]:
        yield ADUser(MO_UUID, ADDate(ENDDATE_FIELD, None))


class _MockADUserSourceMatchingADUserAndEndDate(_MockADUserSource):
    def __iter__(self) -> Iterator[ADUser]:
        yield ADUser(MO_UUID, ADDate(ENDDATE_FIELD, VALID_AD_DATE))


class _MockADUserSourceMatchingADUserWrongEndDate(_MockADUserSource):
    def __iter__(self) -> Iterator[ADUser]:
        yield ADUser(MO_UUID, ADDate(ENDDATE_FIELD, "2023-01-01"))


class _MockADUserSourceMatchingADUserWithSplitEndDates(_MockADUserSource):
    def __iter__(self) -> Iterator[ADUser]:
        yield ADUser(
            MO_UUID,
            ADDate(ENDDATE_FIELD, VALID_AD_DATE),
            ADDate(ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
            None,  # start_date_future
            AD_ORG_UNIT_PATH_EMPTY,
        )


class _TestableChangeListExecutor(ChangeListExecutor):
    def __init__(self):
        super().__init__(settings=TEST_SETTINGS)
        self._ps_scripts_run = []

    def _run_ps_script(self, ps_script):
        self._ps_scripts_run.append(ps_script)
        return {}


class _TestableChangeListExecutorReturningError(_TestableChangeListExecutor):
    def _run_ps_script(self, ps_script):
        super()._run_ps_script(ps_script)
        return {"error": "is mocked"}  # non-empty dict indicates a Powershell error


def dt(val: str):
    parsed: datetime.datetime = datetime.datetime.fromisoformat(val)
    return parsed if parsed.tzinfo else parsed.astimezone(DEFAULT_TIMEZONE)


def validity(start: str, end: str | None) -> dict:
    from_date = dt(start)
    to_date = dt(end) if end else None
    return {
        "validity": {
            "from": str(from_date),
            "to": str(to_date) if to_date else None,
        }
    }


def org_unit() -> dict:
    return {
        "org_unit": [
            {
                "name": "Enhedsnavn",
                "ancestors_validity": [
                    {"name": "Forældreenhed"},
                    {"name": "Topenhed"},
                ],
            }
        ]
    }


def _expected_org_unit_path() -> str:
    return "Topenhed\\Forældreenhed\\Enhedsnavn"


def engagement_objects(*validities: dict) -> dict:
    objects: list[dict] = []
    for validity in validities:
        validity.update(org_unit())
        objects.append(validity)
    return {"engagements": [{"objects": objects}]}


def _get_mock_graphql_session(return_value):
    graphql_session = Mock()
    graphql_session.execute = Mock()
    graphql_session.execute.return_value = return_value
    return graphql_session


@pytest.fixture()
def mock_graphql_session():
    return _get_mock_graphql_session(
        {
            "engagements": [
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2020-01-01",
                                "to": VALID_AD_DATE,
                            },
                            **org_unit(),
                        }
                    ]
                }
            ]
        }
    )


@pytest.fixture()
def mock_graphql_session_raising_keyerror():
    return _get_mock_graphql_session({})


@pytest.fixture()
def mock_mo_engagement_source(mock_graphql_session):
    return MOEngagementSource(mock_graphql_session, split=False)


@pytest.fixture()
def mock_mo_engagement_source_raising_keyerror(
    mock_graphql_session_raising_keyerror,
):
    return MOEngagementSource(mock_graphql_session_raising_keyerror, split=False)


@given(st.builds(Unset), st.builds(Unset))
def test_unset_class_always_equals_itself(unset_a: Unset, unset_b: Unset):
    assert unset_a == unset_b


@given(st.builds(Unset), st.datetimes() | st.none())
def test_unset_class_never_equals_other_types(unset: Unset, other: dict | None):
    assert unset != other


@given(st.builds(Unset))
def test_unset_repr_is_constant(unset: Unset):
    assert repr(unset) == "Unset()"


def test_get_org_unit_path():
    parsed_engagement: _ParsedEngagement = _ParsedEngagement(
        from_dt=None,  # type: ignore
        to_dt=None,  # type: ignore
        org_unit=org_unit()["org_unit"],
    )
    assert parsed_engagement.get_org_unit_path() == _expected_org_unit_path()


def test_mo_engagement_source_raises_keyerror_on_no_engagements():
    mock_mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session({"engagements": []}),
        split=True,
    )
    with pytest.raises(KeyError):
        mock_mo_engagement_source._get_employee_engagements(MO_UUID)


@pytest.mark.parametrize(
    "eng",
    [
        {
            "engagements": [
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2020-09-02T00:00:00+02:00",
                                "to": "2021-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        }
                    ]
                },
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2021-09-02T00:00:00+02:00",
                                "to": "2022-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        }
                    ]
                },
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2022-09-02T00:00:00+02:00",
                                "to": "2023-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        }
                    ]
                },
            ]
        },
        {
            "engagements": [
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2020-09-02T00:00:00+02:00",
                                "to": "2021-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        },
                        {
                            "validity": {
                                "from": "2021-09-02T00:00:00+02:00",
                                "to": "2022-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        },
                        {
                            "validity": {
                                "from": "2022-09-02T00:00:00+02:00",
                                "to": "2023-09-02T00:00:00+02:00",
                            },
                            **org_unit(),
                        },
                    ]
                }
            ]
        },
    ],
)
def test_get_simple_engagement(eng):
    mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(eng),
        split=False,
    )
    expected_latest_date = datetime.date(2023, 9, 2)
    result: MOSimpleEngagement = mo_engagement_source.get_simple_engagement(AD_USER)
    actual_latest_date = result.end_date.date()
    assert actual_latest_date == expected_latest_date


def test_get_simple_engagement_returns_unset_on_no_engagements():
    mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(engagement_objects()),
        split=False,
    )
    assert mo_engagement_source.get_simple_engagement(AD_USER) == MOSimpleEngagement(
        AD_USER, Unset()
    )


def test_date_conversion():
    mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(
            engagement_objects(
                validity(
                    "2005-02-03T00:00:00+01:00",
                    "2038-09-09T00:00:00+02:00",
                )
            )
        ),
        split=False,
    )
    result: MOSplitEngagement = mo_engagement_source.get_split_engagement(AD_USER)
    assert result.end_date.date() == datetime.date(2038, 9, 9)


@pytest.mark.parametrize(
    "mock_response,expected_split",
    [
        # Case A1: only past and present engagements; latest engagement has a blank end
        # date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                PositiveInfinity(),
                Unset(),
                Unset(),
                Unset(),
            ),
        ),
        # Case A2: only past and present engagements; latest engagement has a non-blank
        # end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", "2022-12-31"),
            ),
            MOSplitEngagement(
                AD_USER,
                dt("2022-12-31"),
                Unset(),
                Unset(),
                Unset(),
            ),
        ),
        # Case B1: only future engagements; earliest engagement has a blank end date
        (
            engagement_objects(
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                Unset(),
                PositiveInfinity(),
                dt("2024-01-01"),
                _expected_org_unit_path(),
            ),
        ),
        # Case B2: only future engagements; earliest engagement has an end date
        (
            engagement_objects(
                validity("2024-01-01", "2024-12-31"),
                validity("2025-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                Unset(),
                dt("2024-12-31"),
                dt("2024-01-01"),
                _expected_org_unit_path(),
            ),
        ),
        # Case C1: engagements in the past, present and future. Both the latest
        # engagement in the present, and the earliest engagement in the future, have a
        # blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                PositiveInfinity(),
                PositiveInfinity(),
                dt("2024-01-01"),
                _expected_org_unit_path(),
            ),
        ),
        # Case C2: engagements in the past, present and future. The latest engagement in
        # the present has a blank end date, while the earliest engagement in the future
        # has a non-blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
                validity("2024-01-01", "2024-12-31"),
                validity("2025-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                PositiveInfinity(),
                dt("2024-12-31"),
                dt("2024-01-01"),
                _expected_org_unit_path(),
            ),
        ),
        # Case C3: engagements in the past, present and future. The latest engagement in
        # the present has a non-blank end date, while the earliest engagement in the
        # future has a blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", "2022-12-31"),
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            MOSplitEngagement(
                AD_USER,
                dt("2022-12-31"),
                PositiveInfinity(),
                dt("2024-01-01"),
                _expected_org_unit_path(),
            ),
        ),
    ],
)
def test_get_split_engagement(mock_response, expected_split):
    mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(mock_response),
        split=True,
    )
    with freeze_time("2023-12-01"):
        actual_split: MOSplitEngagement = mo_engagement_source.get_split_engagement(
            AD_USER
        )
    assert actual_split == expected_split


@pytest.mark.parametrize(
    "engagement_objects",
    [
        {"engagements": []},  # no engagements
        engagement_objects(),  # `engagements` key is present but `objects` is empty
    ],
)
def test_get_split_engagement_returns_unset_values_on_no_engagements(
    engagement_objects,
):
    mock_mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(engagement_objects),
        split=True,
    )
    expected_result = MOSplitEngagement(AD_USER, Unset(), Unset(), Unset(), Unset())
    actual_result = mock_mo_engagement_source.get_split_engagement(AD_USER)
    assert actual_result == expected_result


@pytest.mark.parametrize(
    "enddate_field_future,mock_engagements,mock_ad_user_source,expected_result",
    [
        # Case 1a: we are updating only the normal end date in AD, and the user only has
        # one engagement (in the past.) We should write the end date of that engagement
        # to the regular end date field in AD.
        (
            None,
            engagement_objects(validity(VALID_AD_DATE, VALID_AD_DATE)),
            _MockADUserSourceMatchingADUserAndEndDate(),
            [],  # no changes emitted, as MO and AD end dates are already the same
        ),
        # Case 1b: we are updating only the normal end date in AD, and the user only has
        # one engagement (in the past), whose end date is blank. We should write the end
        # date of that engagement to the regular end date field in AD.
        (
            None,
            engagement_objects(validity(VALID_AD_DATE, None)),
            _MockADUserSourceMatchingADUserAndEndDate(),
            [MOSimpleEngagement(AD_USER_VALID_DATE, PositiveInfinity())],
        ),
        # Case 2a: we are updating both the normal and the future end date in AD, and
        # the user only has one engagement (in the past.) We should write the end date
        # of that engagement to the regular end date field in AD, and leave the future
        # end date field in AD unset.
        (
            ENDDATE_FIELD_FUTURE,
            engagement_objects(validity(VALID_AD_DATE, VALID_AD_DATE)),
            _MockADUserSourceMatchingADUserWithSplitEndDates(),
            [],  # no changes emitted, as MO and AD end dates are already the same
        ),
        # Case 2b: we are updating both the normal and the future end date in AD, and
        # the user only has one engagement (in the past), whose end date is blank.  We
        # should write the end date of that engagement to the regular end date field in
        # AD, and leave the future end date field in AD unset.
        (
            ENDDATE_FIELD_FUTURE,
            engagement_objects(validity(VALID_AD_DATE, None)),
            _MockADUserSourceMatchingADUserWithSplitEndDates(),
            [
                MOSplitEngagement(
                    AD_USER_VALID_DATE_FUTURE,
                    PositiveInfinity(),
                    Unset(),
                    Unset(),
                    Unset(),
                ),
            ],
        ),
    ],
)
def test_get_changes(
    enddate_field_future, mock_engagements, mock_ad_user_source, expected_result
):
    mock_mo_engagement_source = MOEngagementSource(
        _get_mock_graphql_session(mock_engagements),
        split=enddate_field_future is not None,
    )
    instance = ChangeList(mock_mo_engagement_source, mock_ad_user_source)
    actual_result = list(instance.get_changes())
    assert actual_result == expected_result


@pytest.mark.parametrize(
    "input_mo_value,expected_return_value",
    [
        # MO and AD values are identical, yield nothing
        (dt(VALID_AD_DATE), None),
        # MO and AD values differ, yield MO value
        (dt("2023-01-01"), dt("2023-01-01")),
        # MO value is None, yield "max date"
        (None, PositiveInfinity().as_datetime()),
        # MO value is positive infinity, yield its datetime value
        (PositiveInfinity(), PositiveInfinity().as_datetime()),
        # MO value is negative infinity, yield its datetime value
        (NegativeInfinity(), NegativeInfinity().as_datetime()),
        # MO value is unset, yield nothing
        (Unset(), None),
    ],
)
def test_compare_end_date(input_mo_value, expected_return_value):
    ad_user = AD_USER_VALID_DATE
    instance = MOSimpleEngagement(ad_user, end_date=input_mo_value)
    actual_return_value = instance._compare_end_date(ad_user.end_date, input_mo_value)
    assert actual_return_value == expected_return_value


@patch("integrations.ad_integration.ad_common.AD._create_session")
@pytest.mark.parametrize(
    "mock_ad_user_source,expected_changes",
    [
        # If no matching AD user, don't return a MO user UUID and MO end date
        (_MockADUserSourceNoMatchingADUser(), None),
        # If matching AD user exists *and* its AD end date is already up to date, don't
        # return a MO user UUID and MO end date.
        (_MockADUserSourceMatchingADUserAndEndDate(), None),
        # If matching AD user exists *but* its AD end date is *blank*, return the MO
        # user UUID and MO end date.
        (_MockADUserSourceMatchingADUser(), {ENDDATE_FIELD: VALID_AD_DATE}),
        # If matching AD user exists *but* its AD end date is *not up to date*, return
        # the MO user UUID and MO end date.
        (
            _MockADUserSourceMatchingADUserWrongEndDate(),
            {ENDDATE_FIELD: VALID_AD_DATE},
        ),
    ],
)
def test_get_changes_simple_engagement(
    mock_create_session,  # patch(...)
    mock_mo_engagement_source,  # pytest fixture
    mock_ad_user_source: ADUserSource,  # pytest parametrize, arg 0
    expected_changes,  # pytest parametrize, arg 1
):
    instance = ChangeList(mock_mo_engagement_source, mock_ad_user_source)
    actual_result = list(instance.get_changes())
    if expected_changes:
        assert len(actual_result) == 1
        assert isinstance(actual_result[0], MOSimpleEngagement)
        assert actual_result[0].ad_user.mo_uuid == MO_UUID
        assert actual_result[0].changes == expected_changes
    else:
        assert len(actual_result) == 0


@patch("integrations.ad_integration.ad_common.AD._create_session")
def test_get_changes_handles_keyerror(
    mock_create_session,
    mock_mo_engagement_source_raising_keyerror,
):
    instance = ChangeList(
        mock_mo_engagement_source_raising_keyerror,
        _MockADUserSourceMatchingADUser(),
    )
    assert list(instance.get_changes()) == []


@pytest.mark.parametrize(
    "reader,enddate_field_future,expected_result,expected_normalized_value",
    [
        (MockADParameterReader(), None, [], "<unused>"),
        (
            MockADParameterReaderWithMOUUID(),
            None,
            [
                ADUser(
                    MO_UUID,
                    ADDate(ENDDATE_FIELD, Invalid()),
                    None,
                    None,
                    ADText(ORG_UNIT_PATH_FIELD_FUTURE, Invalid()),
                )
            ],
            Invalid(),
        ),
        (
            MockADParameterReaderWithMOUUIDAndInvalidDate(),
            None,
            [
                ADUser(
                    MO_UUID,
                    ADDate(ENDDATE_FIELD, INVALID_AD_DATE),
                    None,
                    None,
                    ADText(ORG_UNIT_PATH_FIELD_FUTURE, Invalid()),
                )
            ],
            Invalid(),
        ),
        (
            MockADParameterReaderWithMOUUIDAndValidSplitDates(),
            ENDDATE_FIELD_FUTURE,
            [
                ADUser(
                    MO_UUID,
                    ADDate(ENDDATE_FIELD, VALID_AD_DATE),
                    ADDate(ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
                    None,
                    ADText(ORG_UNIT_PATH_FIELD_FUTURE, Invalid()),
                ),
            ],
            Invalid(),
        ),
    ],
)
def test_ad_user_source(
    reader: ADParameterReader,
    enddate_field_future: str | None,
    expected_result: list[ADUser],
    expected_normalized_value: datetime.datetime | Invalid,
):
    def assert_matches(actual_list: list[ADUser], expected_list: list[ADUser]):
        actual: ADUser
        expected: ADUser
        for actual, expected in zip(actual_list, expected_list):
            assert actual == expected
            assert actual.end_date == expected.end_date
            assert actual.end_date_future == expected.end_date_future
            assert actual.start_date_future == expected.start_date_future
            assert actual.org_unit_path == expected.org_unit_path

    with patch(
        "integrations.ad_integration.ad_fix_enddate.ADParameterReader",
        return_value=reader,
    ):
        instance = ADUserSource(
            AD_UUID_FIELD,
            ENDDATE_FIELD,
            enddate_field_future,
            None,  # startdate_field_future
            ORG_UNIT_PATH_FIELD_FUTURE,
            settings=TEST_SETTINGS,
        )

        assert_matches(list(instance.of_all_users()), expected_result)
        assert_matches(list(instance.of_one_user("username")), expected_result)


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(uuid=st.uuids(), enddate=st.dates())
def test_get_update_cmd(mock_session, uuid, enddate):
    instance = _TestableChangeListExecutor()
    expected_cmd = f"""
    Get-ADUser -SearchBase "{TEST_SEARCH_BASE}" -Credential $usercredential -Filter \'{AD_UUID_FIELD} -eq "{uuid}"\' |
    Set-ADUser -Credential $usercredential -Replace @{{"{ENDDATE_FIELD}"="{enddate}"}} |
    ConvertTo-Json
    """
    actual_cmd = instance.get_update_cmd(
        AD_UUID_FIELD, uuid, **{ENDDATE_FIELD: enddate}
    )
    assert actual_cmd == instance.remove_redundant(expected_cmd)


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(
    st.lists(
        st.builds(MOSimpleEngagement) | st.builds(MOSplitEngagement),
        max_size=10,
    )
)
def test_run_all(mock_session, changes: list):
    instance = _TestableChangeListExecutor()
    retval = instance.run_all(changes, AD_UUID_FIELD)
    for ps_script, ps_result in retval:
        assert "Set-ADUser" in ps_script
        assert ps_result == {}


@patch("integrations.ad_integration.ad_common.AD._create_session")
def test_run_all_logs_results_and_errors(mock_session, caplog):
    mock_changes = [MOSimpleEngagement(AD_USER_VALID_DATE, dt("2022-01-01"))]
    instance = _TestableChangeListExecutorReturningError()
    with caplog.at_level(logging.DEBUG):
        instance.run_all(mock_changes, AD_UUID_FIELD)
    assert len(caplog.records) == 5
    assert "Updating AD user" in caplog.records[0].message
    assert "Set-ADUser" in caplog.records[1].message  # command itself is logged
    assert "AD error response" in caplog.records[2].message  # error from AD is logged
    assert caplog.records[3].message == "0 users end dates corrected"
    assert caplog.records[4].message == "All end dates are fixed"


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(
    st.lists(
        st.builds(MOSimpleEngagement) | st.builds(MOSplitEngagement),
        min_size=1,
        max_size=1,
    )
)
def test_run_all_supports_dry_run(mock_session, changes: list):
    instance = _TestableChangeListExecutor()
    retval = instance.run_all(changes, AD_UUID_FIELD, dry=True)
    if retval:
        ps_script, ps_result = retval[0]
        assert "Set-ADUser" in ps_script
        assert ps_result == "<dry run>"
    assert len(instance._ps_scripts_run) == 0


def mock_option_default(option, context, **kwargs):
    if option.name == "auth_server":
        return "http://keycloak"
    return None


@patch("click.core.Option.get_default", new=mock_option_default)
@patch("integrations.ad_integration.ad_common.read_settings", return_value={})
@patch("integrations.ad_integration.ad_common.AD._create_session")
@patch(
    "integrations.ad_integration.ad_fix_enddate.ADUserSource",
    cls=_MockADUserSourceNoMatchingADUser,
)
@pytest.mark.parametrize(
    "args",
    [
        [],
        ["--ad-user", "foobar"],
        ["--dry-run"],
    ],
)
def test_cli(
    mock_read_settings,
    mock_create_session,
    mock_ad_user_source,
    args,
    caplog,
):
    runner = CliRunner()
    with caplog.at_level(logging.INFO):
        result = runner.invoke(cli, args)
    if result.exception is not None:
        raise result.exception
    assert caplog.records[0].message.startswith("Command line args: ")
    assert caplog.records[-1].message == "All end dates are fixed"



import uuid
from unittest import TestCase
from unittest.mock import patch

import pytest

from ..ad_writer import EngagementDatesError
from ..ad_writer import LoraCacheSource
from ..ad_writer import MOGraphqlSource
from ..ad_writer import MORESTSource
from ..utils import AttrDict
from .mocks import MockMORESTSource


class TestMOGraphqlSource:
    def _get_instance(self, *objs) -> MOGraphqlSource:
        wrapped_value = [{"objects": [obj]} for obj in objs]
        with patch(
            "integrations.ad_integration.ad_writer.MOGraphqlSource._run_query",
            return_value=wrapped_value,
        ):
            return MOGraphqlSource({})

    def _get_mock_engagement(self, employee_uuid, is_primary=True, end=None):
        return {
            "employee_uuid": employee_uuid,
            "is_primary": is_primary,
            "validity": {"to": end.isoformat() if end else None},
        }

    def _get_manager_uuid(self, instance, employee_uuid):
        return instance.get_manager_uuid({"uuid": employee_uuid}, None)

    def test_employee_has_manager_in_same_org_unit(self):
        employee_uuid = str(uuid.uuid4())
        manager_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            {
                "uuid": str(uuid.uuid4()),  # org unit UUID
                "parent_uuid": None,
                "engagements": [self._get_mock_engagement(employee_uuid)],
                "managers": [{"employee_uuid": manager_uuid}],
            }
        )
        assert self._get_manager_uuid(instance, employee_uuid) == manager_uuid

    def test_employee_is_self_manager(self):
        org_unit_uuid = str(uuid.uuid4())
        parent_org_unit_uuid = str(uuid.uuid4())
        employee_uuid = str(uuid.uuid4())
        manager_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            # Child org unit: where the employee is engaged and also is a "self manager"
            {
                "uuid": org_unit_uuid,
                "parent_uuid": parent_org_unit_uuid,
                "engagements": [self._get_mock_engagement(employee_uuid)],
                "managers": [{"employee_uuid": employee_uuid}],  # "self manager"
            },
            # Parent org unit: where the "real" manager is defined
            {
                "uuid": parent_org_unit_uuid,
                "parent_uuid": None,
                "engagements": [],
                "managers": [{"employee_uuid": manager_uuid}],  # the "real" manager
            },
        )
        assert self._get_manager_uuid(instance, employee_uuid) == manager_uuid

    def test_employee_is_self_manager_no_parent_managers(self):
        employee_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            {
                "uuid": str(uuid.uuid4()),  # org unit UUID
                "parent_uuid": None,
                "engagements": [self._get_mock_engagement(employee_uuid)],
                "managers": [{"employee_uuid": employee_uuid}],
            }
        )
        assert self._get_manager_uuid(instance, employee_uuid) is None

    def test_employee_has_no_primary_engagement(self):
        employee_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            {
                "uuid": str(uuid.uuid4()),  # org unit UUID
                "parent_uuid": None,
                "engagements": [
                    self._get_mock_engagement(employee_uuid, is_primary=None)
                ],
                "managers": [{"employee_uuid": str(uuid.uuid4())}],
            }
        )
        assert self._get_manager_uuid(instance, employee_uuid) is None

    def test_employee_has_no_engagements(self):
        employee_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            {
                "uuid": str(uuid.uuid4()),  # org unit UUID
                "parent_uuid": None,
                "engagements": [],
                "managers": [{"employee_uuid": str(uuid.uuid4())}],
            }
        )
        assert self._get_manager_uuid(instance, employee_uuid) is None

    def test_employee_has_no_managers(self):
        employee_uuid = str(uuid.uuid4())
        instance = self._get_instance(
            {
                "uuid": str(uuid.uuid4()),  # org unit UUID
                "parent_uuid": None,
                "engagements": [self._get_mock_engagement(employee_uuid)],
                "managers": [],
            }
        )
        assert self._get_manager_uuid(instance, employee_uuid) is None

    def test_follows_pagination(self):
        """Verify that `MOGraphqlSource._run_query` follows the pagination cursors in
        the responses returned by the GraphQL API.
        """

        class _MockClient:
            """Mock GraphQL client whose `execute` method simulates a paginated set of
            results.
            """

            def __init__(self):
                self.num_pages = 5
                self.current_page = 0

            def __enter__(self):
                # Implement the context manager protocol, returning the instance itself
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                # Implement the context manager protocol,
                pass

            def execute(self, query, **kwargs):
                response = {
                    "org_units": {
                        "page_info": {
                            "next_cursor": (
                                str(self.current_page)
                                if self.current_page < self.num_pages
                                else None
                            )
                        },
                        "objects": [],
                    }
                }
                if self.current_page < self.num_pages:
                    self.current_page += 1
                return response

        # Arrange: test `MOGraphqlSource` using our mock GraphQL client
        mock_client = _MockClient()
        with patch(
            "integrations.ad_integration.ad_writer.MOGraphqlSource._get_client",
            return_value=mock_client,
        ):
            instance = MOGraphqlSource({})
            # Act: consume the paginated result set
            instance._run_query()
            # Assert: check that we reached the last page of the paginated result set.
            # (We assume that we also visited the other pages in the result set.)
            assert mock_client.current_page == mock_client.num_pages


class TestLoraCacheSource(TestCase):
    def setUp(self):
        self.user = self.setup_user()
        self.lc = AttrDict(
            {
                "users": {
                    self.user["uuid"]: [self.user],
                },
                "engagements": {
                    "key-1": [
                        # Current registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2020-01-01",
                            "to_date": None,
                        },
                        # Previous registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2019-01-01",
                            "to_date": "2021-01-01",
                        },
                    ],
                },
            }
        )
        self.lc_historic = AttrDict(
            {
                "engagements": {
                    # Add some historic engagements that must not be read by
                    # `get_engagement_dates`.
                    "key-1": [
                        # Current historic registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2010-01-01",
                            "to_date": "2019-12-31",
                        },
                    ]
                }
            }
        )

    def setup_user(self):
        return {
            "uuid": "some_uuid_here",
            "navn": "some_name some_lastname",
            "efternavn": "some_lastname",
            "fornavn": "some_name",
            "kaldenavn": "some_alias some_lastname_alias",
            "kaldenavn_fornavn": "some_alias",
            "kaldenavn_efternavn": "some_lastname_alias",
            "cpr": "some_cpr",
        }

    def test_read_user(self):
        datasource = self._get_datasource(None, None)
        result = datasource.read_user("some_uuid_here")
        self.assertEqual(
            result,
            {
                "uuid": "some_uuid_here",
                "name": "some_name some_lastname",
                "surname": "some_lastname",
                "givenname": "some_name",
                "nickname": "some_alias some_lastname_alias",
                "nickname_givenname": "some_alias",
                "nickname_surname": "some_lastname_alias",
                "cpr_no": "some_cpr",
            },
        )

    def test_get_engagement_dates(self):
        datasource = self._get_datasource("2020-01-01", None)
        result = datasource.get_engagement_dates(self.user["uuid"])
        self.assertEqual(
            [list(elem) for elem in result],  # consume each iterable in result
            [["2020-01-01"], [None]],
        )

    def test_get_engagement_endpoint_dates(self):
        datasource = self._get_datasource("2020-01-01", None)
        result = datasource.get_engagement_endpoint_dates(self.user["uuid"])
        # "to_date" of None must be converted into "9999-12-31"
        self.assertEqual(result, ("2020-01-01", "9999-12-31"))

    def _get_datasource(self, from_date, to_date):
        return LoraCacheSource(
            self.lc, self.lc_historic, MockMORESTSource(from_date, to_date)
        )


class TestEngagementDates:
    MOSource = MORESTSource(settings={"global": {"mora.base": "http://lol.money:1337"}})

    def test_valid_interval(self, monkeypatch):
        start, end = "2021-01-01", "2021-02-02"

        def _valid_dates(*args):
            return [start], [end]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _valid_dates)
        assert self.MOSource.get_engagement_endpoint_dates("test") == (start, end)

    def test_sentinels(self, monkeypatch):
        start, end = "1930-01-01", "9999-12-31"

        def _falsy_dates(*args):
            return [None], [""]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _falsy_dates)
        assert self.MOSource.get_engagement_endpoint_dates("test") == (start, end)

    def test_invalid_dates(self, monkeypatch):
        start, end = "2021-12-31", "2021-01-01"

        def _invalid_dates(*args):
            return [start], [end]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _invalid_dates)
        with pytest.raises(EngagementDatesError, match=f"{start}.*{end}"):
            self.MOSource.get_engagement_endpoint_dates("test")

        def _empty_dates(*args):
            return [], []

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _empty_dates)
        with pytest.raises(EngagementDatesError, match="9999-12-31.*1930-01-01"):
            self.MOSource.get_engagement_endpoint_dates("test")



import json
import logging
import uuid
from typing import Dict
from typing import Optional
from typing import Tuple

import click
import sentry_sdk
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm

from .ad_exceptions import CprNotFoundInADException
from .ad_exceptions import CprNotNotUnique
from .ad_exceptions import ManagerNotUniqueFromCprException
from .ad_exceptions import UserNotFoundException
from .ad_logger import start_logging
from .ad_reader import ADParameterReader
from .ad_writer import ADWriter
from exporters.sql_export.lora_cache import fetch_loracache


logger = logging.getLogger("MoAdSync")
export_logger = logging.getLogger("export")


def run_mo_to_ad_sync(
    reader: ADParameterReader,
    writer: ADWriter,
    mo_uuid_field: str,
    sync_cpr: Optional[str] = None,
    sync_username: Optional[str] = None,
    dry_run: Optional[bool] = False,
):
    if sync_cpr or sync_username:
        print("Warning: --sync-cpr/--sync-username is for testing only")
        all_users = [reader.read_user(user=sync_username, cpr=sync_cpr)]
    else:
        all_users = reader.read_it_all(print_progress=True)

    def filter_missing_uuid_field(user):
        if mo_uuid_field.lower() not in set(k.lower() for k in user):
            msg = "User {} does not have a {} field - skipping"
            logger.info(msg.format(user["SamAccountName"], mo_uuid_field))
            return False
        return True

    def update_stats(stats: Dict[str, int], response) -> Dict[str, int]:
        if response[0]:
            stats["fully_synced"] += 1
            if response[1] == "Sync completed":
                stats["updated"] += 1
                if response[2] is False:
                    stats["no_manager"] += 1

            if response[1] == "Nothing to edit":
                stats["nothing_to_edit"] += 1
                if response[2] is False:
                    stats["no_manager"] += 1
        else:
            if response[1] == "No active engagments":
                stats["no_active_engagement"] += 1
            else:
                stats["unknown_failed_sync"] += 1
        return stats

    stats = {
        "attempted_users": 0,
        "fully_synced": 0,
        "nothing_to_edit": 0,
        "updated": 0,
        "no_manager": 0,
        "unknown_manager_failure": 0,
        "cpr_not_unique": 0,
        "user_not_in_mo": 0,
        "user_not_in_ad": 0,
        "critical_error": 0,
        "unknown_failed_sync": 0,
        "no_active_engagement": 0,
    }

    all_users = list(filter(filter_missing_uuid_field, all_users))
    logger.info("Will now attempt to sync {} users".format(len(all_users)))

    for user in tqdm(all_users, unit="user"):
        if dry_run:
            stats["attempted_users"] += 1
            mo_uuid = user[mo_uuid_field]
            mo_values = writer.read_ad_information_from_mo(
                mo_uuid, ad_dump=all_users, read_manager=False
            )
            if mo_values:
                logger.info(
                    "Would update AD user %r (MO user = %r, location = %r)",
                    user["SamAccountName"],
                    mo_uuid,
                    mo_values["location"],
                )
                stats["updated"] += 1
            else:
                logger.info(
                    "SKIP update of AD user %r (MO user = %r)",
                    user["SamAccountName"],
                    mo_uuid,
                )
                stats["nothing_to_edit"] += 1
            continue

        stats["attempted_users"] += 1
        msg = "Now syncing: {}, {}".format(user["SamAccountName"], user[mo_uuid_field])
        logger.info(msg)
        try:
            response = writer.sync_user(user[mo_uuid_field], ad_dump=all_users)
            logger.debug("Respose to sync: {}".format(response))
            stats = update_stats(stats, response)
        except ManagerNotUniqueFromCprException:
            stats["unknown_manager_failure"] += 1
            msg = "Did not find a unique manager for {}".format(user[mo_uuid_field])
            logger.error(msg)
        except CprNotNotUnique:
            stats["cpr_not_unique"] += 1
            msg = "User {} with uuid: {} has more than one AD account"
            logger.error(msg.format(user["Name"], user[mo_uuid_field]))
        except CprNotFoundInADException:
            stats["user_not_in_ad"] += 1
            msg = "User {}, {} with uuid {} could not be found by cpr"
            logger.error(
                msg.format(user["SamAccountName"], user["Name"], user[mo_uuid_field])
            )
        except UserNotFoundException:
            stats["user_not_in_mo"] += 1
            msg = "User {}, {} with uuid {} was not found i MO, unable to sync"
            logger.error(
                msg.format(user["SamAccountName"], user["Name"], user[mo_uuid_field])
            )
        except Exception as e:
            stats["critical_error"] += 1
            logger.error("Unhandled exception: {}".format(e))
            logger.exception("Unhandled exception:")
            export_logger.error(
                "Error updating AD user %r: %s", user["SamAccountName"], e
            )
            print("Unhandled exception: {}".format(e))

    print()
    print(json.dumps(stats, indent=4))
    logger.info("Stats: {}".format(stats))

    return stats


def run_preview_command_for_uuid(
    reader: ADParameterReader,
    writer: ADWriter,
    mo_uuid: uuid.UUID,
    sync_username: Optional[str] = None,
) -> Tuple[str]:
    mo_values = writer.read_ad_information_from_mo(mo_uuid)

    ad_dump = [
        # Retrieve AD user data for employee
        reader.read_user(user=sync_username),
        # Retrieve AD user data for manager, if available
        (
            reader.read_user(cpr=mo_values["manager_cpr"])
            if mo_values.get("manager_cpr")
            else {}
        ),
    ]

    (
        sync_cmd,
        rename_cmd,
        rename_cmd_target,
        add_manager_cmd,
    ) = writer._preview_sync_command(mo_uuid, sync_username, ad_dump=ad_dump)

    click.echo_via_pager(sync_cmd)
    click.echo_via_pager(rename_cmd)
    click.echo_via_pager(f"Rename targets AD user: {rename_cmd_target!r}")
    click.echo_via_pager(add_manager_cmd)

    return sync_cmd, rename_cmd, rename_cmd_target, add_manager_cmd  # type: ignore


@click.command()
@click.option(
    "--lora-speedup/--no-lora-speedup",
    help="Utilize LoraCache to speedup the operation",
    is_flag=True,
    default=lambda: load_settings()["integrations.ad_writer.lora_speedup"],
)
@click.option(
    "--mo-uuid-field",
    type=click.STRING,
    default=lambda: load_settings()["integrations.ad.write.uuid_field"],
)
@click.option(
    "--sync-cpr",
    help="Synchronize the specified user instead of all users",
    type=click.STRING,
)
@click.option(
    "--sync-username",
    help="Synchronize the specified user instead of all users",
    type=click.STRING,
)
@click.option("--ignore-occupied-names", is_flag=True, default=False)
@click.option(
    "--preview-command-for-uuid",
    help="Given a MO user UUID, preview the PowerShell command(s) to run",
    type=click.STRING,
)
@click.option(
    "--dry-run",
    help="Run dry updates (only verifies if AD user would be updated)",
    is_flag=True,
    default=False,
)
def main(
    lora_speedup: bool,
    mo_uuid_field: str,
    sync_cpr: Optional[str],
    sync_username: Optional[str],
    ignore_occupied_names: bool,
    preview_command_for_uuid: Optional[uuid.UUID],
    dry_run: bool,
):
    start_logging()

    settings = load_settings()
    if "crontab.SENTRY_DSN" in settings:
        sentry_sdk.init(dsn=settings["crontab.SENTRY_DSN"])

    reader = ADParameterReader()

    lc, lc_historic = fetch_loracache() if lora_speedup else (None, None)
    writer = ADWriter(
        lc=lc,
        lc_historic=lc_historic,
        # XXX: occupied_names should not be an empty array, but it takes
        # forever to initialize, essentially reading all of AD.
        # TODO: We should support on-demand name generation without pre-seed.
        skip_occupied_names=ignore_occupied_names,
    )

    if preview_command_for_uuid:
        if sync_cpr:
            raise click.ClickException("you cannot use --sync-cpr with this option")
        if not sync_username:
            raise click.ClickException(
                "please specify --sync-username (samAccountName)"
            )
        run_preview_command_for_uuid(
            reader,
            writer,
            preview_command_for_uuid,
            sync_username=sync_username,
        )
        return

    run_mo_to_ad_sync(
        reader,
        writer,
        mo_uuid_field,
        sync_cpr=sync_cpr,
        sync_username=sync_username,
        dry_run=dry_run,
    )


if __name__ == "__main__":
    main()
